import pytest
from pytest import raises

import numpy as np
import netket as nk

from .. import common

pytestmark = common.skipif_mpi

SEED = 214748364


def _setup_system():
    L = 3

    hi = nk.hilbert.Spin(s=0.5) ** L

    ha = nk.operator.LocalOperator(hi)
    j_ops = []
    for i in range(L):
        ha += (0.3 / 2.0) * nk.operator.spin.sigmax(hi, i)
        ha += (
            (2.0 / 4.0)
            * nk.operator.spin.sigmaz(hi, i)
            * nk.operator.spin.sigmaz(hi, (i + 1) % L)
        )
        j_ops.append(nk.operator.spin.sigmam(hi, i))

    #  Create the liouvillian
    lind = nk.operator.LocalLiouvillian(ha, j_ops)
    return hi, lind


def _setup_ss(dtype=np.float32, sr=True):
    hi, lind = _setup_system()

    ma = nk.models.NDM()
    # sa = nk.sampler.ExactSampler(hilbert=nk.hilbert.DoubledHilber(hi), n_chains=16)

    sa = nk.sampler.MetropolisLocal(hilbert=nk.hilbert.DoubledHilbert(hi))
    sa_obs = nk.sampler.MetropolisLocal(hilbert=hi)

    vs = nk.vqs.MCMixedState(sa, ma, sampler_diag=sa_obs, n_samples=1000, seed=SEED)

    op = nk.optimizer.Sgd(learning_rate=0.05)
    if sr:
        sr_config = nk.optimizer.SR()
    else:
        sr_config = None

    driver = nk.SteadyState(lind, op, variational_state=vs, preconditioner=sr_config)

    return lind, vs, driver


def _setup_obs(L):
    hi = nk.hilbert.Spin(s=0.5) ** L

    obs_sx = nk.operator.LocalOperator(hi)
    for i in range(L):
        obs_sx += nk.operator.spin.sigmax(hi, i)

    obs = {"SigmaX": obs_sx}
    return obs


####


def test_estimate():
    lind, _, driver = _setup_ss()

    driver.estimate(lind.H @ lind)
    driver.advance(1)
    driver.estimate(lind.H @ lind)


def test_raise_n_iter():
    lind, _, driver = _setup_ss()
    with raises(
        ValueError,
    ):
        driver.run("prova", 12)


def test_steadystate_construction_vstate():
    lind, vs, driver = _setup_ss()

    sa = vs.sampler
    op = nk.optimizer.Sgd(learning_rate=0.05)

    driver = nk.SteadyState(lind, op, sa, nk.models.NDM(), n_samples=1000, seed=SEED)

    driver.run(1)

    assert driver.step_count == 1

    with raises(TypeError):
        ha2 = nk.operator.LocalOperator(lind.hilbert_physical)
        driver = nk.SteadyState(ha2, op, variational_state=driver.state)


def test_steadystate_steadystate_legacy_api():
    lind, vs, driver = _setup_ss()
    op = driver.optimizer
    vs = driver.state
    sr_config = driver.preconditioner

    with pytest.raises(ValueError):
        driver = nk.SteadyState(
            lind, op, variational_state=vs, sr=sr_config, preconditioner=sr_config
        )

    driver = nk.SteadyState(
        lind, op, variational_state=vs, sr=sr_config, sr_restart=True
    )
    assert driver.preconditioner == sr_config



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
from flax.linen.initializers import *
from jax import numpy as jnp
from netket.jax.utils import dtype_real


def _compute_fans(shape, in_axis=-2, out_axis=-1):
    receptive_field_size = shape.total / shape[in_axis] / shape[out_axis]
    fan_in = shape[in_axis] * receptive_field_size
    fan_out = shape[out_axis] * receptive_field_size
    return fan_in, fan_out


def _complex_uniform(key, shape, dtype):
    """
    Sample uniform random values within a disk on the complex plane,
    with zero mean and unit variance.
    """
    key_r, key_theta = jax.random.split(key)
    dtype = dtype_real(dtype)
    r = jnp.sqrt(2 * jax.random.uniform(key_r, shape, dtype))
    theta = 2 * jnp.pi * jax.random.uniform(key_theta, shape, dtype)
    return r * jnp.exp(1j * theta)


def _complex_truncated_normal(key, upper, shape, dtype):
    """
    Sample random values from a centered normal distribution on the complex plane,
    whose modulus is truncated to `upper`, and the variance before the truncation is one.
    """
    key_r, key_theta = jax.random.split(key)
    dtype = dtype_real(dtype)
    t = (1 - jnp.exp(-(upper ** 2))) * jax.random.uniform(key_r, shape, dtype)
    r = jnp.sqrt(-jnp.log(1 - t))
    theta = 2 * jnp.pi * jax.random.uniform(key_theta, shape, dtype)
    return r * jnp.exp(1j * theta)


def variance_scaling(
    scale, mode, distribution, in_axis=-2, out_axis=-1, dtype=jnp.float32
):
    """
    Initializer capable of adapting its scale to the shape of the weights tensor.

    With `distribution="truncated_normal" or "normal"`, samples are
    drawn from a truncated/untruncated normal distribution with a mean of zero and
    a standard deviation (after truncation, if used) `stddev = sqrt(scale / n)`,
    where `n` is:
    - number of input units in the weights tensor, if `mode="fan_in"`
    - number of output units, if `mode="fan_out"`
    - average of the numbers of input and output units, if `mode="fan_avg"`

    With `distribution="truncated_normal"`, the absolute values of the samples are
    truncated below 2 standard deviations before truncation.

    With `distribution="uniform"`, samples are drawn from:
    - a uniform interval, if `dtype` is real
    - a uniform disk, if `dtype` is complex
    with a mean of zero and a standard deviation of `stddev`.

    The generalization of variance scaling initializers to complex numbers is discussed in,
    e.g., `Trabelsi et. {\\it al} <https://arxiv.org/abs/1705.09792>`_.
    They proposed an axisymmetric distribution implemented by first sampling the modulus
    from the radial CDF, then uniformly sampling the phase.

    Args:
      scale: scaling factor (positive float).
      mode: one of "fan_in", "fan_out", and "fan_avg".
      distribution: random distribution to use. One of "truncated_normal",
        "normal" and "uniform".
      in_axis: axis of the input dimension in the weights tensor (default: -2).
      out_axis: axis of the output dimension in the weights tensor (default: -1).
      dtype: the dtype of the weights (default: float32).
    """

    def init(key, shape, dtype=dtype):
        shape = jax.core.as_named_shape(shape)
        fan_in, fan_out = _compute_fans(shape, in_axis, out_axis)
        if mode == "fan_in":
            denominator = fan_in
        elif mode == "fan_out":
            denominator = fan_out
        elif mode == "fan_avg":
            denominator = (fan_in + fan_out) / 2
        else:
            raise ValueError(
                "invalid mode for variance scaling initializer: {}".format(mode)
            )

        variance = jnp.array(scale / denominator, dtype=dtype)

        if distribution == "truncated_normal":
            if jnp.issubdtype(dtype, jnp.floating):
                # constant is stddev of standard normal truncated to (-2, 2)
                stddev = jnp.sqrt(variance) / 0.87962566103423978
                return jax.random.truncated_normal(key, -2, 2, shape, dtype) * stddev
            else:
                stddev = jnp.sqrt(variance) / 0.95311164380491208
                return _complex_truncated_normal(key, 2, shape, dtype) * stddev
        elif distribution == "normal":
            return jax.random.normal(key, shape, dtype) * jnp.sqrt(variance)
        elif distribution == "uniform":
            if jnp.issubdtype(dtype, jnp.floating):
                stddev = jnp.sqrt(3 * variance)
                return jax.random.uniform(key, shape, dtype, -1) * stddev
            else:
                return _complex_uniform(key, shape, dtype) * jnp.sqrt(variance)
        else:
            raise ValueError("invalid distribution for variance scaling initializer")

    return init


glorot_uniform = partial(variance_scaling, 1, "fan_avg", "uniform")
glorot_normal = partial(variance_scaling, 1, "fan_avg", "truncated_normal")
lecun_uniform = partial(variance_scaling, 1, "fan_in", "uniform")
lecun_normal = partial(variance_scaling, 1, "fan_in", "truncated_normal")
he_uniform = partial(variance_scaling, 2, "fan_in", "uniform")
he_normal = partial(variance_scaling, 2, "fan_in", "truncated_normal")

xavier_uniform = glorot_uniform
xavier_normal = glorot_normal
kaiming_uniform = he_uniform
kaiming_normal = he_normal



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netket as nk

import jax.numpy as jnp
import jax.random as random
import numpy as np
import scipy.sparse
from jax.lax import dot
from netket.utils.group import PermutationGroup

import pytest


def _setup_symm(symmetries, N, lattice=nk.graph.Chain):

    g = lattice(N)

    hi = nk.hilbert.Spin(1 / 2, g.n_nodes)

    if symmetries == "trans":
        # Only translations, N_symm = N_sites
        perms = g.translation_group()
    else:
        # All chain automorphisms, N_symm = 2 N_sites
        perms = g.space_group()

    return g, hi, perms


@pytest.mark.parametrize("symmetries", ["trans", "space_group"])
@pytest.mark.parametrize("use_bias", [True, False])
@pytest.mark.parametrize("mode", ["fft", "matrix"])
def test_DenseSymm(symmetries, use_bias, mode):
    rng = nk.jax.PRNGSeq(0)

    g, hi, perms = _setup_symm(symmetries, N=8)

    if mode == "matrix":
        ma = nk.nn.DenseSymm(
            symmetries=perms,
            mode=mode,
            features=8,
            use_bias=use_bias,
            bias_init=nk.nn.initializers.uniform(),
        )
    else:
        ma = nk.nn.DenseSymm(
            symmetries=perms,
            shape=tuple(g.extent),
            mode=mode,
            features=8,
            use_bias=use_bias,
            bias_init=nk.nn.initializers.uniform(),
        )

    pars = ma.init(rng.next(), hi.random_state(rng.next(), 1))

    v = hi.random_state(rng.next(), 3)
    vals = [ma.apply(pars, v[..., p]) for p in np.asarray(perms)]
    for val in vals:
        assert jnp.allclose(jnp.sum(val, -1), jnp.sum(vals[0], -1))


@pytest.mark.parametrize("mode", ["fft", "matrix", "irreps"])
def test_DenseEquivariant_creation(mode):
    g = nk.graph.Chain(8)
    space_group = g.space_group()
    hi = nk.hilbert.Spin(1 / 2, N=8)

    def check_init(creator):
        ma = creator()
        _ = ma.init(nk.jax.PRNGKey(0), np.ones([1, 4, 16]))

    perms = [[0, 1, 2, 3, 4, 5, 6, 7], [7, 6, 5, 4, 3, 2, 1, 0]]

    # Init with graph
    check_init(
        lambda: nk.nn.DenseEquivariant(
            symmetries=g,
            mode=mode,
            in_features=4,
            out_features=4,
        )
    )

    # init with space_group
    if mode == "irreps":
        check_init(
            lambda: nk.nn.DenseEquivariant(
                symmetries=space_group,
                mode=mode,
                in_features=4,
                out_features=4,
            )
        )
    else:
        check_init(
            lambda: nk.nn.DenseEquivariant(
                symmetries=space_group,
                shape=tuple(g.extent),
                mode=mode,
                in_features=4,
                out_features=4,
            )
        )

    # init with arrays
    if mode == "irreps":
        check_init(
            lambda: nk.nn.DenseEquivariant(
                symmetries=space_group.irrep_matrices(),
                mode=mode,
                in_features=4,
                out_features=4,
            )
        )
    elif mode == "fft":
        check_init(
            lambda: nk.nn.DenseEquivariant(
                symmetries=space_group.product_table,
                mode=mode,
                shape=(8,),
                in_features=4,
                out_features=4,
            )
        )
    else:
        check_init(
            lambda: nk.nn.DenseEquivariant(
                symmetries=space_group.product_table,
                mode=mode,
                in_features=4,
                out_features=4,
            )
        )


@pytest.mark.parametrize("symmetries", ["trans", "space_group"])
@pytest.mark.parametrize("use_bias", [True, False])
@pytest.mark.parametrize("lattice", [nk.graph.Chain, nk.graph.Square])
@pytest.mark.parametrize("mode", ["fft", "matrix", "irreps"])
@pytest.mark.parametrize("mask", [True, False])
def test_DenseEquivariant(symmetries, use_bias, lattice, mode, mask):
    rng = nk.jax.PRNGSeq(0)

    g, hi, perms = _setup_symm(symmetries, N=3, lattice=lattice)

    pt = perms.product_table
    n_symm = np.asarray(perms).shape[0]

    if mask:
        mask = np.random.randint(0, 2, [n_symm])
    else:
        mask = np.ones([n_symm])

    if mode == "irreps":
        ma = nk.nn.DenseEquivariant(
            symmetries=perms,
            mode=mode,
            in_features=1,
            out_features=1,
            mask=mask,
            use_bias=use_bias,
            bias_init=nk.nn.initializers.uniform(),
        )
    else:
        ma = nk.nn.DenseEquivariant(
            symmetries=pt,
            shape=tuple(g.extent),
            mode=mode,
            in_features=1,
            out_features=1,
            mask=mask,
            use_bias=use_bias,
            bias_init=nk.nn.initializers.uniform(),
        )

    pars = ma.init(rng.next(), np.random.normal(0, 1, [1, 1, n_symm]))

    # inv_pt computes chosen_op = gh^-1 instead of g^-1h
    chosen_op = np.random.randint(n_symm)
    inverse = PermutationGroup(
        [perms.elems[i] for i in perms.inverse], degree=g.n_nodes
    )
    inv_pt = inverse.product_table
    sym_op = np.where(inv_pt == chosen_op, 1.0, 0.0)

    v = random.normal(rng.next(), [3, 1, n_symm])
    v_trans = jnp.matmul(v, sym_op)

    out = ma.apply(pars, v)
    out_trans = ma.apply(pars, v_trans)

    # output should be involution
    assert jnp.allclose(jnp.matmul(out, sym_op), out_trans)


@pytest.mark.parametrize("lattice", [nk.graph.Chain, nk.graph.Square])
@pytest.mark.parametrize("symmetries", ["trans", "space_group"])
def test_modes_DenseSymm(lattice, symmetries):

    rng = nk.jax.PRNGSeq(0)
    g, hi, perms = _setup_symm(symmetries, N=3, lattice=lattice)

    ma_fft = nk.nn.DenseSymm(
        symmetries=perms,
        mode="fft",
        features=4,
        shape=tuple(g.extent),
        bias_init=nk.nn.initializers.uniform(),
    )
    ma_matrix = nk.nn.DenseSymm(
        symmetries=perms,
        mode="matrix",
        features=4,
        bias_init=nk.nn.initializers.uniform(),
    )

    dum_input = np.random.normal(0, 1, [1, g.n_nodes])
    pars = ma_fft.init(rng.next(), dum_input)
    _ = ma_matrix.init(rng.next(), dum_input)

    assert jnp.allclose(ma_fft.apply(pars, dum_input), ma_matrix.apply(pars, dum_input))


@pytest.mark.parametrize("lattice", [nk.graph.Chain, nk.graph.Square])
@pytest.mark.parametrize("symmetries", ["trans", "space_group"])
def test_modes_DenseEquivariant(lattice, symmetries):

    rng = nk.jax.PRNGSeq(0)
    g, hi, perms = _setup_symm(symmetries, N=3, lattice=lattice)

    ma_fft = nk.nn.DenseEquivariant(
        symmetries=perms,
        mode="fft",
        in_features=1,
        out_features=1,
        shape=tuple(g.extent),
        bias_init=nk.nn.initializers.uniform(),
    )
    ma_irreps = nk.nn.DenseEquivariant(
        symmetries=perms,
        mode="irreps",
        in_features=1,
        out_features=1,
        bias_init=nk.nn.initializers.uniform(),
    )
    ma_matrix = nk.nn.DenseEquivariant(
        symmetries=perms,
        mode="matrix",
        in_features=1,
        out_features=1,
        bias_init=nk.nn.initializers.uniform(),
    )

    dum_input = np.random.normal(0, 1, [1, 1, len(perms)])
    pars = ma_fft.init(rng.next(), dum_input)
    _ = ma_irreps.init(rng.next(), dum_input)
    _ = ma_matrix.init(rng.next(), dum_input)

    fft_out = ma_fft.apply(pars, dum_input)
    irreps_out = ma_irreps.apply(pars, dum_input)
    matrix_out = ma_matrix.apply(pars, dum_input)

    assert jnp.allclose(fft_out, irreps_out)
    assert jnp.allclose(fft_out, matrix_out)


@pytest.mark.parametrize("symmetries", ["trans", "space_group"])
@pytest.mark.parametrize("features", [1, 2, 5])
def test_symmetrizer(symmetries, features):
    from netket.nn.symmetric_linear import _symmetrizer_col

    _, _, perms = _setup_symm(symmetries, N=8)

    n_symm, n_sites = perms.shape
    n_hidden = features * n_symm

    # symmetrization tensor entries
    def symmetrizer_ijkl(i, j, k, l):
        jsymm = np.floor_divide(j, n_symm)
        cond_k = k == np.asarray(perms)[j % n_symm, i]
        cond_l = l == jsymm
        return np.asarray(np.logical_and(cond_k, cond_l), dtype=int)

    symmetrizer = np.asarray(
        np.fromfunction(
            symmetrizer_ijkl,
            shape=(n_sites, n_hidden, n_sites, features),
            dtype=np.intp,
        ),
    ).reshape(-1, features * n_sites)
    symmetrizer = scipy.sparse.coo_matrix(symmetrizer)

    # Of the COO matrix attributes, rows is just a range [0, ..., n_rows)
    # and data is [1., ..., 1.]. Only cols is non-trivial.
    assert np.all(symmetrizer.row == np.arange(symmetrizer.shape[0]))
    assert np.all(symmetrizer.data == 1.0)
    assert np.all(symmetrizer.col == _symmetrizer_col(np.asarray(perms), features))



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from functools import partial, wraps
from netket.jax import compose

import jax
import jax.flatten_util
import jax.numpy as jnp

import numpy as np

from netket.stats import subtract_mean
from netket.utils import mpi

from netket.utils.types import Array, Callable, PyTree, Scalar

from netket.jax import tree_cast, tree_conj, tree_axpy, tree_to_real


# TODO better name and move it somewhere sensible
def single_sample(forward_fn):
    """
    A decorator to make the forward_fn accept a single sample
    """

    def f(W, σ):
        return forward_fn(W, σ[jnp.newaxis, :])[0]

    return f


# TODO move it somewhere reasonable
def tree_subtract_mean(oks: PyTree) -> PyTree:
    """
    subtract the mean with MPI along axis 0 of every leaf
    """
    return jax.tree_map(partial(subtract_mean, axis=0), oks)  # MPI


@partial(jax.vmap, in_axes=(None, None, 0))
def jacobian_real_holo(forward_fn: Callable, params: PyTree, samples: Array) -> PyTree:
    """Calculates Jacobian entries by vmapping grad.
    Assumes the function is R→R or holomorphic C→C, so single grad is enough

    Args:
        forward_fn: the log wavefunction ln Ψ
        params : a pytree of parameters p
        samples : an array of n samples σ

    Returns:
        The Jacobian matrix ∂/∂pₖ ln Ψ(σⱼ) as a PyTree
    """
    y, vjp_fun = jax.vjp(single_sample(forward_fn), params, samples)
    res, _ = vjp_fun(np.array(1.0, dtype=jnp.result_type(y)))
    return res


@partial(jax.vmap, in_axes=(None, None, 0, None))
def _jacobian_cplx(
    forward_fn: Callable, params: PyTree, samples: Array, _build_fn: Callable
) -> PyTree:
    """Calculates Jacobian entries by vmapping grad.
    Assumes the function is R→C, backpropagates 1 and -1j

    Args:
        forward_fn: the log wavefunction ln Ψ
        params : a pytree of parameters p
        samples : an array of n samples σ

    Returns:
        The Jacobian matrix ∂/∂pₖ ln Ψ(σⱼ) as a PyTree
    """
    y, vjp_fun = jax.vjp(single_sample(forward_fn), params, samples)
    gr, _ = vjp_fun(np.array(1.0, dtype=jnp.result_type(y)))
    gi, _ = vjp_fun(np.array(-1.0j, dtype=jnp.result_type(y)))
    return _build_fn(gr, gi)


@partial(wraps(_jacobian_cplx))
def jacobian_cplx(
    forward_fn, params, samples, _build_fn=partial(jax.tree_multimap, jax.lax.complex)
):
    return _jacobian_cplx(forward_fn, params, samples, _build_fn)


centered_jacobian_real_holo = compose(tree_subtract_mean, jacobian_real_holo)
centered_jacobian_cplx = compose(tree_subtract_mean, jacobian_cplx)


def _divide_by_sqrt_n_samp(oks, samples):
    """
    divide Oⱼₖ by √n
    """
    n_samp = samples.shape[0] * mpi.n_nodes  # MPI
    return jax.tree_map(lambda x: x / np.sqrt(n_samp), oks)


def stack_jacobian(centered_oks: PyTree) -> PyTree:
    """
    Return the real and imaginary parts of ΔOⱼₖ stacked along the sample axis
    Re[S] = Re[(ΔOᵣ + i ΔOᵢ)ᴴ(ΔOᵣ + i ΔOᵢ)] = ΔOᵣᵀ ΔOᵣ + ΔOᵢᵀ ΔOᵢ = [ΔOᵣ ΔOᵢ]ᵀ [ΔOᵣ ΔOᵢ]
    """
    return jax.tree_map(
        lambda x: jnp.concatenate([x.real, x.imag], axis=0), centered_oks
    )


def stack_jacobian_tuple(centered_oks_re_im):
    """
    stack the real and imaginary parts of ΔOⱼₖ along the sample axis

    Re[S] = Re[(ΔOᵣ + i ΔOᵢ)ᴴ(ΔOᵣ + i ΔOᵢ)] = ΔOᵣᵀ ΔOᵣ + ΔOᵢᵀ ΔOᵢ = [ΔOᵣ ΔOᵢ]ᵀ [ΔOᵣ ΔOᵢ]

    Args:
        centered_oks_re_im : a tuple (ΔOᵣ, ΔOᵢ) of two PyTrees representing the real and imag part of ΔOⱼₖ
    """
    return jax.tree_multimap(
        lambda re, im: jnp.concatenate([re, im], axis=0), *centered_oks_re_im
    )


def _rescale(centered_oks):
    """
    compute ΔOₖ/√Sₖₖ and √Sₖₖ
    to do scale-invariant regularization (Becca & Sorella 2017, pp. 143)
    Sₖₗ/(√Sₖₖ√Sₗₗ) = ΔOₖᴴΔOₗ/(√Sₖₖ√Sₗₗ) = (ΔOₖ/√Sₖₖ)ᴴ(ΔOₗ/√Sₗₗ)
    """
    scale = jax.tree_map(
        lambda x: mpi.mpi_sum_jax(jnp.sum((x * x.conj()).real, axis=0, keepdims=True))[
            0
        ]
        ** 0.5,
        centered_oks,
    )
    centered_oks = jax.tree_multimap(jnp.divide, centered_oks, scale)
    scale = jax.tree_map(partial(jnp.squeeze, axis=0), scale)
    return centered_oks, scale


def _jvp(oks: PyTree, v: PyTree) -> Array:
    """
    Compute the matrix-vector product between the pytree jacobian oks and the pytree vector v
    """
    td = lambda x, y: jnp.tensordot(x, y, axes=y.ndim)
    return jax.tree_util.tree_reduce(jnp.add, jax.tree_multimap(td, oks, v))


def _vjp(oks: PyTree, w: Array) -> PyTree:
    """
    Compute the vector-matrix product between the vector w and the pytree jacobian oks
    """
    res = jax.tree_map(partial(jnp.tensordot, w, axes=1), oks)
    return jax.tree_map(lambda x: mpi.mpi_sum_jax(x)[0], res)  # MPI


def _mat_vec(v: PyTree, oks: PyTree) -> PyTree:
    """
    Compute ⟨O† O⟩v = ∑ₗ ⟨Oₖᴴ Oₗ⟩ vₗ
    """
    res = tree_conj(_vjp(oks, _jvp(oks, v).conjugate()))
    return tree_cast(res, v)


# ==============================================================================
# the logic above only works for R→R, R→C and holomorphic C→C
# here the other modes are converted


@partial(jax.jit, static_argnums=(0, 4, 5))
def prepare_centered_oks(
    apply_fun: Callable,
    params: PyTree,
    samples: Array,
    model_state: Optional[PyTree],
    mode: str,
    rescale_shift: bool,
) -> PyTree:
    """
    compute ΔOⱼₖ = Oⱼₖ - ⟨Oₖ⟩ = ∂/∂pₖ ln Ψ(σⱼ) - ⟨∂/∂pₖ ln Ψ⟩
    divided by √n

    In a somewhat intransparent way this also internally splits all parameters to real
    in the 'real' and 'complex' modes (for C→R, R&C→R, R&C→C and general C→C) resulting in the respective ΔOⱼₖ
    which is only compatible with split-to-real pytree vectors

    Args:
        apply_fun: The forward pass of the Ansatz
        params : a pytree of parameters p
        samples : an array of (n in total) batched samples σ
        model_state: untrained state parameters of the model
        mode: differentiation mode, must be one of 'real', 'complex', 'holomorphic'
        rescale_shift: whether scale-invariant regularisation should be used (default: True)

    Returns:
        if not rescale_shift:
            a pytree representing the centered jacobian of ln Ψ evaluated at the samples σ, divided by √n;
            None
        else:
            the same pytree, but the entries for each parameter normalised to unit norm;
            pytree containing the norms that were divided out (same shape as params)

    """
    # un-batch the samples
    samples = samples.reshape((-1, samples.shape[-1]))

    # pre-apply the model state
    def forward_fn(W, σ):
        return apply_fun({"params": W, **model_state}, σ)

    if mode == "real":
        split_complex_params = True  # convert C→R and R&C→R to R→R
        centered_jacobian_fun = centered_jacobian_real_holo
    elif mode == "complex":
        split_complex_params = True  # convert C→C and R&C→C to R→C
        # centered_jacobian_fun = compose(stack_jacobian, centered_jacobian_cplx)

        # avoid converting to complex and then back
        # by passing around the oks as a tuple of two pytrees representing the real and imag parts
        centered_jacobian_fun = compose(
            stack_jacobian_tuple,
            partial(centered_jacobian_cplx, _build_fn=lambda *x: x),
        )
    elif mode == "holomorphic":
        split_complex_params = False
        centered_jacobian_fun = centered_jacobian_real_holo
    else:
        raise NotImplementedError(
            'Differentiation mode should be one of "real", "complex", or "holomorphic", got {}'.format(
                mode
            )
        )

    if split_complex_params:
        # doesn't do anything if the params are already real
        params, reassemble = tree_to_real(params)

        def f(W, σ):
            return forward_fn(reassemble(W), σ)

    else:
        f = forward_fn

    centered_oks = _divide_by_sqrt_n_samp(
        centered_jacobian_fun(
            f,
            params,
            samples,
        ),
        samples,
    )
    if rescale_shift:
        return _rescale(centered_oks)
    else:
        return centered_oks, None


def mat_vec(v: PyTree, centered_oks: PyTree, diag_shift: Scalar) -> PyTree:
    """
    Compute (S + δ) v = 1/n ⟨ΔO† ΔO⟩v + δ v = ∑ₗ 1/n ⟨ΔOₖᴴΔOₗ⟩ vₗ + δ vₗ

    Only compatible with R→R, R→C, and holomorphic C→C
    for C→R, R&C→R, R&C→C and general C→C the parameters for generating ΔOⱼₖ should be converted to R,
    and thus also the v passed to this function as well as the output are expected to be of this form

    Args:
        v: pytree representing the vector v compatible with centered_oks
        centered_oks: pytree of gradients 1/√n ΔOⱼₖ
        diag_shift: a scalar diagonal shift δ
    Returns:
        a pytree corresponding to the sr matrix-vector product (S + δ) v
    """
    return tree_axpy(diag_shift, v, _mat_vec(v, centered_oks))



from netket import legacy as nk
import flax
from flax import linen as nn
import jax
from jax import numpy as jnp
import numpy as np


# 1D Lattice
L = 20
g = nk.graph.Hypercube(length=L, n_dim=1, pbc=True)

# Hilbert space of spins on the graph
hi = nk.hilbert.Spin(s=0.5) ** L

ha = nk.operator.Ising(h=1.0, hilbert=hi, graph=g)

alpha = 1


class RBM(nn.Module):
    @nn.compact
    def __call__(self, x):
        x = nn.Dense(features=alpha * x.shape[-1], dtype=jax.numpy.float32)(x)
        x = jnp.log(jnp.cosh(x))
        return jnp.sum(x, axis=-1)


ma = nk.machine.Flax(hi, RBM())
ma.init_random_parameters(seed=1232)

# Jax Sampler
sa = nk.sampler.MetropolisLocal(machine=ma, n_chains=2)

# Using Sgd
op = nk.optimizer.Sgd(ma, learning_rate=0.1)

# Create the optimization driver
gs = nk.Vmc(
    hamiltonian=ha, sampler=sa, optimizer=op, n_samples=1000, sr=None, n_discard=None
)

# The first iteration is slower because of start-up jit times
gs.run(out="test", n_iter=2)

gs.run(out="test", n_iter=300)



# Copyright 2020 The Netket Authors. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from netket import legacy as nk
from ed import load_ed_data

L = 10
J2 = 0.4

# Load the Hilbert space info and data
hi, training_samples, training_targets = load_ed_data(L, J2)

# Machine
ma = nk.machine.RbmSpin(hilbert=hi, alpha=1)
ma.init_random_parameters(seed=1234, sigma=0.01)

# Optimizer
op = nk.optimizer.AdaDelta()


spvsd = nk.supervised.Supervised(
    machine=ma,
    optimizer=op,
    n_chains=400,
    samples=training_samples,
    targets=training_targets,
)

n_iter = 4000

# Run with "Overlap_phi" loss. Also available currently is "MSE, Overlap_uni"
spvsd.run(
    n_iter=n_iter,
    loss_function="Overlap_phi",
    out="output",
    save_params_every=50,
)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket.utils.dispatch import parametric

from ._abstract_operator import AbstractOperator


class WrappedOperator(AbstractOperator):
    def get_conn_flattened(self, *args):
        return NotImplementedError

    def __add__(self, other):
        return self.collect() + other

    def __radd__(self, other):
        return other + self.collect()

    def __sub__(self, other):
        return self.collect() - other

    def __rsub__(self, other):
        return other - self.collect()

    def __mul__(self, other):
        return self.collect() * other

    def __rmul__(self, other):
        return other * self.collect()

    def _op__matmul__(self, other):
        return self.collect() @ other

    def _op__rmatmul__(self, other):
        return other @ self.collect()


@parametric
class Transpose(WrappedOperator):
    """
    A wrapper lazily representing the transpose of the wrapped object.
    """

    parent: AbstractOperator
    """The wrapped object"""

    def __init__(self, op):
        super().__init__(op.hilbert)
        self.parent = op

    @property
    def dtype(self):
        return self.parent.dtype

    def collect(self):
        return self.parent.transpose(concrete=True)

    def transpose(self, *, concrete=False):
        return self.parent

    def conjugate(self, *, concrete=False):
        if self.parent.is_hermitian:
            return self.parent

        if concrete:
            return self.parent.conjugate(concrete=True)

        return Adjoint(self.parent)

    @property
    def H(self):
        return self.parent.conjugate()

    def __repr__(self):
        return "Transpose({})".format(self.parent)

    def get_conn_flattened(self, *args):
        return NotImplementedError

    def to_sparse(self):
        return self.parent.to_sparse().T


@parametric
class Adjoint(WrappedOperator):
    """
    A wrapper lazily representing the adjoint (conjugate-transpose) of the wrapped object.
    """

    parent: AbstractOperator
    """The wrapped object"""

    def __init__(self, op):
        super().__init__(op.hilbert)
        self.parent = op

    @property
    def dtype(self):
        return self.parent.dtype

    def collect(self):
        return self.parent.transpose(concrete=True).conj(concrete=True)

    def transpose(self, *, concrete=False):
        return self.parent.conjugate(concrete=concrete)

    def conjugate(self, concrete=False):
        return self.parent.transpose(concrete=concrete)

    @property
    def H(self):
        return self.parent

    def __repr__(self):
        return "Adjoint({})".format(self.parent)

    def __mul__(self, other):
        if self.parent == other:
            return Squared(other)

        return self.collect() * other

    def __rmul__(self, other):
        if self.parent == other:
            return Squared(other.H.collect())

        return other * self.collect()

    def _op__matmul__(self, other):
        if self.parent == other:
            return Squared(other)

        return self.collect() @ other

    # def __op_rmatmul__(self, other):
    #    if self.parent == other:
    #        return Squared(other.H.collect())
    #
    #    return other @ self.collect()

    def to_sparse(self):
        return self.parent.to_sparse().T.conj()


@parametric
class Squared(WrappedOperator):
    """
    A wrapper lazily representing the matrix-squared (A^dag A) of the wrapped object A.
    """

    parent: AbstractOperator
    """The wrapped object"""

    def __init__(self, op):
        super().__init__(op.hilbert)
        self.parent = op

    @property
    def dtype(self):
        return self.parent.dtype

    def collect(self):
        return self.parent.H.collect()._concrete_matmul_(self.parent)

    @property
    def T(self):
        self.parent.c @ self.parent.T
        self.parent
        return self.parent.conjugate()

    def conjugate(self):
        return Transpose(self.parent)

    @property
    def H(self):
        return self

    @property
    def is_hermitian(self) -> bool:
        return True

    def __repr__(self):
        return "Squared({})".format(self.parent)

    def __mul__(self, other):
        return self.collect() * other

    def __rmul__(self, other):
        return other * self.collect()

    def _op__matmul__(self, other):
        return self.collect() @ other

    def to_sparse(self):
        op = self.parent.to_sparse()
        return op.conj().T @ op



from numbers import Number

from typing import Any

from .dispatch import dispatch
from .types import Array


@dispatch
def dtype(x: Number):
    return type(x)


@dispatch
def dtype(x: Array):  # noqa: F811, E0102
    return x.dtype


@dispatch
def is_scalar(_: Any):
    return False


@dispatch
def is_scalar(_: Number):  # noqa: F811, E0102
    return True


@dispatch
def is_scalar(x: Array):  # noqa: F811, E0102
    return x.ndim == 0



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netket as nk
import numpy as np
import optax
import pytest
from jax import numpy as jnp

from .. import common

pytestmark = common.skipif_mpi


@pytest.mark.parametrize("dtype", [jnp.float64, jnp.complex128])
@pytest.mark.parametrize(
    "hilbert",
    [
        pytest.param(
            nk.hilbert.Spin(s=1 / 2, N=4),
            id="spin_1/2",
        ),
        pytest.param(
            nk.hilbert.Spin(s=1, N=4),
            id="spin_1",
        ),
        pytest.param(
            nk.hilbert.Fock(n_max=3, N=4),
            id="fock",
        ),
    ],
)
@pytest.mark.parametrize(
    "partial_models",
    [
        # pytest.param(
        #     (
        #         lambda hilbert, dtype: nk.models.ARNNDense(
        #             hilbert=hilbert,
        #             layers=3,
        #             features=5,
        #             dtype=dtype,
        #         ),
        #         lambda hilbert, dtype: nk.models.FastARNNDense(
        #             hilbert=hilbert,
        #             layers=3,
        #             features=5,
        #             dtype=dtype,
        #         ),
        #     ),
        #     id="dense",
        # ),
        pytest.param(
            (
                lambda hilbert, dtype: nk.models.ARNNConv1D(
                    hilbert=hilbert,
                    layers=3,
                    features=5,
                    kernel_size=2,
                    dtype=dtype,
                ),
                lambda hilbert, dtype: nk.models.FastARNNConv1D(
                    hilbert=hilbert,
                    layers=3,
                    features=5,
                    kernel_size=2,
                    dtype=dtype,
                ),
            ),
            id="conv1d",
        ),
        pytest.param(
            (
                lambda hilbert, dtype: nk.models.ARNNConv1D(
                    hilbert=hilbert,
                    layers=3,
                    features=5,
                    kernel_size=2,
                    kernel_dilation=2,
                    dtype=dtype,
                ),
                lambda hilbert, dtype: nk.models.FastARNNConv1D(
                    hilbert=hilbert,
                    layers=3,
                    features=5,
                    kernel_size=2,
                    kernel_dilation=2,
                    dtype=dtype,
                ),
            ),
            id="conv1d_dilation",
        ),
        pytest.param(
            (
                lambda hilbert, dtype: nk.models.ARNNConv2D(
                    hilbert=hilbert,
                    layers=3,
                    features=5,
                    kernel_size=(2, 3),
                    dtype=dtype,
                ),
                lambda hilbert, dtype: nk.models.FastARNNConv2D(
                    hilbert=hilbert,
                    layers=3,
                    features=5,
                    kernel_size=(2, 3),
                    dtype=dtype,
                ),
            ),
            id="conv2d",
        ),
        pytest.param(
            (
                lambda hilbert, dtype: nk.models.ARNNConv2D(
                    hilbert=hilbert,
                    layers=3,
                    features=5,
                    kernel_size=(2, 3),
                    kernel_dilation=(2, 2),
                    dtype=dtype,
                ),
                lambda hilbert, dtype: nk.models.FastARNNConv2D(
                    hilbert=hilbert,
                    layers=3,
                    features=5,
                    kernel_size=(2, 3),
                    kernel_dilation=(2, 2),
                    dtype=dtype,
                ),
            ),
            id="conv2d_dilation",
        ),
    ],
)
def test_AR_VMC(partial_models, hilbert, dtype):
    model1 = partial_models[0](hilbert, dtype)
    model2 = partial_models[1](hilbert, dtype)

    sampler1 = nk.sampler.ARDirectSampler(hilbert, n_chains=3)
    vstate1 = nk.vqs.MCState(sampler1, model1, n_samples=6, seed=123, sampler_seed=456)
    assert vstate1.n_discard_per_chain == 0
    samples1 = vstate1.sample()

    graph = nk.graph.Hypercube(length=hilbert.size, n_dim=1)
    H = nk.operator.Ising(hilbert=hilbert, graph=graph, h=1)
    optimizer = optax.adam(learning_rate=1e-3)
    vmc1 = nk.VMC(H, optimizer, variational_state=vstate1)
    vmc1.run(n_iter=3)
    samples_trained1 = vstate1.sample()

    sampler2 = nk.sampler.ARDirectSampler(hilbert, n_chains=3)
    vstate2 = nk.vqs.MCState(sampler2, model2, n_samples=6, seed=123, sampler_seed=456)
    samples2 = vstate2.sample()

    # Samples from FastARNN* should be the same as those from ARNN*
    np.testing.assert_allclose(samples2, samples1)

    vmc2 = nk.VMC(H, optimizer, variational_state=vstate2)
    vmc2.run(n_iter=3)
    samples_trained2 = vstate2.sample()

    # Samples from FastARNN* after training should be the same as those from ARNN*
    np.testing.assert_allclose(samples_trained2, samples_trained1)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import netket as nk

from .. import common

pytestmark = common.skipif_mpi


def test_deprecated_name():
    with pytest.warns(FutureWarning):
        nk.optim.sr

    with pytest.raises(AttributeError):
        nk.optim.accabalubba

    assert dir(nk.optim) == dir(nk.optimizer)


def test_deprecated_sr():
    with pytest.warns(FutureWarning):
        nk.optimizer.sr.SRLazyCG()

    with pytest.warns(FutureWarning):
        nk.optimizer.sr.SRLazyGMRES()



import netket as nk
from netket.optimizer import qgt
from flax.core import unfreeze
from flax import linen as nn
from flax.linen.initializers import normal, variance_scaling
import jax
from jax import numpy as jnp
from jax.scipy.sparse.linalg import cg
import numpy as np
from typing import Any
from netket.utils.types import NNInitFunc
import timeit

from sys import argv


def timeit_gc(x, number=1):
    return timeit.Timer(x, "gc.enable()").timeit(number=number)


# Return the proper kind of QGT
def QGT(*args, **kwargs):
    if argv[1] == "pytree":
        return qgt.QGTJacobianPyTree(*args, **kwargs)
    elif argv[1] == "onthefly":
        return qgt.QGTOnTheFly(*args, **kwargs)
    else:
        raise Exception("bad kind of QGT")


def construct_and_solve(vstate, rhs):
    qgt_ = QGT(vstate=vstate, diag_shift=0.01)
    return jax.tree_map(lambda x: x.block_until_ready(), qgt_.solve(cg, rhs))


# Benchmark starts here
def benchmark(side, n_samples, layers, features):
    n_nodes = side * side
    keys = jax.random.split(jax.random.PRNGKey(0), 5)

    graph = nk.graph.Square(side)
    hilbert = nk.hilbert.Spin(s=0.5, N=n_nodes)

    machine = nk.models.GCNN(
        symmetries=graph, features=features, layers=layers, dtype=float, mode=argv[2]
    )

    # Create a variational state to run QGT on
    sa = nk.sampler.MetropolisExchange(
        hilbert=hilbert, graph=graph, d_max=2, n_chains=n_samples
    )
    vstate = nk.variational.MCState(
        sampler=sa, model=machine, n_samples=n_samples, n_discard_per_chain=2
    )
    vstate.init(seed=0)
    Tsamp_1 = timeit_gc(lambda: vstate.samples.block_until_ready())

    # Generate a random RHS of the same pytree shape as the parameters
    vec, unravel = nk.jax.tree_ravel(vstate.parameters)
    vec1 = jax.random.normal(keys[1], shape=vec.shape, dtype=vec.dtype)
    rhs1 = unravel(vec1)
    vec2 = jax.random.normal(keys[2], shape=vec.shape, dtype=vec.dtype)
    rhs2 = unravel(vec2)
    # Generate a new set of parameters for the second set of runs
    vecp = vec * jax.random.normal(keys[3], shape=vec.shape, dtype=vec.dtype)
    pars = unravel(vecp)

    Tjit = timeit_gc(
        lambda: construct_and_solve(vstate, rhs1),
        number=1,
    )

    # See what jit hath wrought us
    vstate._parameters = pars
    Tsamp_2 = timeit_gc(lambda: vstate.samples.block_until_ready())
    Tsamp = (Tsamp_1 + Tsamp_2) / 2

    # Time of full SR cycle
    Tcycle = (
        timeit_gc(
            lambda: construct_and_solve(vstate, rhs2),
            number=5,
        )
        / 5
    )

    # Time of solving only
    qgt_ = QGT(vstate=vstate, diag_shift=0.01)

    Tsolve = (
        timeit_gc(
            lambda: jax.tree_map(lambda x: x.block_until_ready(), qgt_.solve(cg, rhs2)),
            number=5,
        )
        / 5
    )

    print(
        f"{side}\t{features}\t{layers}\t{n_samples}\t{Tsamp:.6f}\t{Tjit:.6f}\t{Tcycle:.6f}\t{Tsolve:.6f}"
    )


print(f"# GCNN mode {argv[2]}")
print(f"# Side length\tFeatures\tLayers\tSamples\tSampling\tJitting\tSR cycle\tSolve")

# Different system sizes
benchmark(6, 256, 4, 4)
benchmark(8, 256, 4, 4)
benchmark(10, 256, 4, 4)
benchmark(12, 256, 4, 4)

# Different feature count
benchmark(6, 256, 4, 2)
benchmark(6, 256, 4, 4)
benchmark(6, 256, 4, 6)
benchmark(6, 256, 4, 8)

# Different number of layers
benchmark(6, 256, 4, 4)
benchmark(6, 256, 6, 4)
benchmark(6, 256, 8, 4)
benchmark(6, 256, 10, 4)

# Different sample numbers
benchmark(6, 128, 4, 4)
benchmark(6, 256, 4, 4)
benchmark(6, 384, 4, 4)
benchmark(6, 512, 4, 4)
benchmark(6, 768, 4, 4)
benchmark(6, 1024, 4, 4)



import numpy as np
import matplotlib.pyplot as plt
import json

plt.ion()

while True:
    plt.clf()

    iters = []
    log_overlap = []
    mse = []
    mse_log = []

    data = json.load(open("output.log"))
    for iteration in data["Output"]:
        iters.append(iteration["Iteration"])
        log_overlap.append(iteration["log_overlap"])
        mse.append(iteration["mse"])
        mse_log.append(iteration["mse_log"])

    overlap = np.exp(-np.array(log_overlap))

    J2 = 0.4
    plt.subplot(2, 1, 1)
    plt.title(r"$J_1 J_2$ model, $J_2=" + str(J2) + "$")
    plt.ylabel("Overlap = F")
    plt.xlabel("Iteration #")

    plt.plot(iters, overlap)
    plt.axhline(
        y=1, xmin=0, xmax=iters[-1], linewidth=2, color="k", label="max accuracy = 1"
    )

    plt.legend(frameon=False)

    plt.subplot(2, 1, 2)
    plt.ylabel("Overlap Error = 1-F")
    plt.xlabel("Iteration #")
    plt.semilogy(iters, 1.0 - overlap)

    plt.pause(1)

    plt.show()



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from numba import jit

import numpy as np
from flax import struct

from netket.operator import AbstractOperator


from ..metropolis import MetropolisRule


@struct.dataclass
class CustomRuleState:
    sections: np.ndarray
    rand_op_n: np.ndarray
    weight_cumsum: np.ndarray


@struct.dataclass
class CustomRuleNumpy(MetropolisRule):
    operator: Any = struct.field(pytree_node=False)
    weight_list: Any = struct.field(pytree_node=False, default=None)

    def __post_init__(self):
        # Raise errors if hilbert is not an Hilbert
        if not isinstance(self.operator, AbstractOperator):
            raise TypeError(
                "Argument to CustomRuleNumpy must be a valid operator, "
                f"but operator is a {type(self.operator)}."
            )

        _check_operators(self.operator.operators)

        if self.weight_list is not None:
            if self.weight_list.shape != (self.operator.n_operators,):
                raise ValueError("move_weights have the wrong shape")
            if self.weight_list.min() < 0:
                raise ValueError("move_weights must be positive")
        else:
            object.__setattr__(
                self,
                "weight_list",
                np.ones(self.operator.n_operators, dtype=np.float32),
            )

        object.__setattr__(
            self, "weight_list", self.weight_list / self.weight_list.sum()
        )

    def init_state(rule, sampler, machine, params, key):
        return CustomRuleState(
            sections=np.empty(sampler.n_batches, dtype=np.int32),
            rand_op_n=np.empty(sampler.n_batches, dtype=np.int32),
            weight_cumsum=rule.weight_list.cumsum(),
        )

    def transition(rule, sampler, machine, parameters, state, rng, σ):
        rule_state = state.rule_state

        # numba does not support jitting np.random number generators
        # so we have to generate the random numbers outside the jit
        # block
        rnd_uniform = rng.uniform(0.0, 1.0, size=σ.shape[0])

        _pick_random_and_init(
            σ.shape[0],
            rule_state.weight_cumsum,
            rnd_uniform=rnd_uniform,
            out=rule_state.rand_op_n,
        )

        σ_conns, mels = rule.operator.get_conn_filtered(
            state.σ, rule_state.sections, rule_state.rand_op_n
        )

        # numba does not support jitting np.random number generators
        # so we have to generate the random numbers outside the jit
        # block
        rnd_uniform = rng.uniform(0.0, 1.0, size=state.σ1.shape[0])

        _choose_and_return(
            state.σ1,
            σ_conns,
            mels,
            rule_state.sections,
            state.log_prob_corr,
            rnd_uniform,
        )


@jit(nopython=True)
def _pick_random_and_init(batch_size, move_cumulative, rnd_uniform, out):
    for i in range(batch_size):
        p = rnd_uniform[i]
        out[i] = np.searchsorted(move_cumulative, p)

    # return out


@jit(nopython=True)
def _choose_and_return(σp, x_prime, mels, sections, log_prob_corr, rnd_uniform):
    low = 0
    for i in range(σp.shape[0]):
        p = rnd_uniform[i]
        exit_state = 0
        cumulative_prob = mels[low].real
        while p > cumulative_prob:
            exit_state += 1
            cumulative_prob += mels[low + exit_state].real
        σp[i] = x_prime[low + exit_state]
        low = sections[i]

    log_prob_corr.fill(0.0)


def _check_operators(operators):
    for op in operators:
        assert op.imag.max() < 1.0e-10
        assert op.min() >= 0
        assert np.allclose(op.sum(axis=0), 1.0)
        assert np.allclose(op.sum(axis=1), 1.0)
        assert np.allclose(op, op.T)



from functools import partial

import numpy as np
from jax import numpy as jnp
from netket.utils import get_afun_if_module
from netket.utils import mpi
import jax


def to_array(hilbert, apply_fun, variables, normalize=True):
    """
    Computes `apply_fun(variables, states)` on all states of `hilbert` and returns
      the results as a vector.

    Args:
        normalize: If True, the vector is normalized to have L2-norm 1.
    """
    if not hilbert.is_indexable:
        raise RuntimeError("The hilbert space is not indexable")

    apply_fun = get_afun_if_module(apply_fun)

    # mpi4jax does not have (yet) allgatherv so we need to be creative
    # could be made easier if we update mpi4jax
    n_states = hilbert.n_states
    n_states_padded = int(np.ceil(n_states / mpi.n_nodes)) * mpi.n_nodes
    states_n = np.arange(n_states)
    fake_states_n = np.arange(n_states_padded - n_states)

    # divide the hilbert space in chunks for each node
    states_per_rank = np.split(np.concatenate([states_n, fake_states_n]), mpi.n_nodes)

    xs = hilbert.numbers_to_states(states_per_rank[mpi.rank])

    return _to_array_rank(apply_fun, variables, xs, n_states, normalize)


@partial(jax.jit, static_argnums=(0, 3, 4))
def _to_array_rank(apply_fun, variables, σ_rank, n_states, normalize):
    """
    Computes apply_fun(variables, σ_rank) and gathers all results across all ranks.
    The input σ_rank should be a slice of all states in the hilbert space of equal
    length across all ranks because mpi4jax does not support allgatherv (yet).

    Args:
        n_states: total number of elements in the hilbert space.
    """
    # number of 'fake' states, in the last rank.
    n_fake_states = σ_rank.shape[0] * mpi.n_nodes - n_states

    log_psi_local = apply_fun(variables, σ_rank)

    # last rank, get rid of fake elements
    if mpi.rank == mpi.n_nodes - 1 and n_fake_states > 0:
        log_psi_local = jax.ops.index_update(
            log_psi_local, jax.ops.index[-n_fake_states:], -jnp.inf
        )

    if normalize:
        # subtract logmax for better numerical stability
        logmax, _ = mpi.mpi_max_jax(log_psi_local.real.max())
        log_psi_local -= logmax

    psi_local = jnp.exp(log_psi_local)

    if normalize:
        # compute normalization
        norm2 = jnp.linalg.norm(psi_local) ** 2
        norm2, _ = mpi.mpi_sum_jax(norm2)
        psi_local /= jnp.sqrt(norm2)

    psi, _ = mpi.mpi_allgather_jax(psi_local)
    psi = psi.reshape(-1)

    # remove fake states
    psi = psi[0:n_states]
    return psi


def to_matrix(hilbert, machine, params, normalize=True):

    if not hilbert.is_indexable:
        raise RuntimeError("The hilbert space is not indexable")

    psi = to_array(hilbert, machine, params, normalize=False)

    L = hilbert.physical.n_states
    rho = psi.reshape((L, L))
    if normalize:
        trace = jnp.trace(rho)
        rho /= trace

    return rho



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numba.experimental import jitclass
from numba import int64, float64


spec = [
    ("_local_states", float64[:]),
    ("_local_size", int64),
    ("_size", int64),
    ("_basis", int64[:]),
]


@jitclass(spec)
class HilbertIndex:
    def __init__(self, local_states, size):
        self._local_states = np.sort(local_states)
        self._local_size = len(self._local_states)
        self._size = size

        self._basis = np.zeros(size, dtype=np.int64)
        ba = 1
        for s in range(size):
            self._basis[s] = ba
            ba *= self._local_size

    def _local_state_number(self, x):
        return np.searchsorted(self.local_states, x)

    @property
    def n_states(self):
        return self._local_size ** self._size

    @property
    def local_states(self):
        return self._local_states

    def state_to_number(self, state):
        if state.ndim != 1:
            raise RuntimeError("Invalid input shape, expecting a 1d array.")

        # Converts a vector of quantum numbers into the unique integer identifier
        number = 0

        for i in range(self._size):
            number += (
                self._local_state_number(state[self._size - i - 1]) * self._basis[i]
            )

        return number

    def number_to_state(self, number, out=None):

        if out is None:
            out = np.empty(self._size)
        # else:
        #     assert out.size == self._size

        out.fill(self._local_states[0])

        ip = number
        k = self._size - 1
        while ip > 0:
            out[k] = self._local_states[ip % self._local_size]
            ip = ip // self._local_size
            k -= 1

        return out

    def states_to_numbers(self, states, out=None):
        if states.ndim != 2:
            raise RuntimeError("Invalid input shape, expecting a 2d array.")

        if out is None:
            out = np.empty(states.shape[0], np.int64)
        # else:
        #     assert out.size == states.shape[0]

        for i in range(states.shape[0]):
            out[i] = 0
            for j in range(self._size):
                out[i] += (
                    self._local_state_number(states[i, self._size - j - 1])
                    * self._basis[j]
                )
        return out

    def numbers_to_states(self, numbers, out=None):
        if numbers.ndim != 1:
            raise RuntimeError("Invalid input shape, expecting a 1d array.")

        if out is None:
            out = np.empty((numbers.shape[0], self._size))
        # else:
        #     assert out.shape == (numbers.shape[0], self._size)

        for i in range(numbers.shape[0]):
            out[i] = self.number_to_state(numbers[i])

        return out

    def all_states(self, out=None):
        if out is None:
            out = np.empty((self.n_states, self._size))

        for i in range(self.n_states):
            self.number_to_state(i, out[i])
        return out



from netket.nn import DenseSymm, DenseEquivariant
from netket.graph import Square
import numpy as np
import jax
from timeit import timeit

# Test DenseSymm

for L in [2, 4, 6, 8, 10]:
    dum_input = np.random.normal(0, 1, [1000, L * L])
    graph = Square(L)
    ma_fft = DenseSymm(symmetries=graph, mode="fft", features=4)
    ma_matrix = DenseSymm(symmetries=graph, mode="matrix", features=4)

    params_fft = ma_fft.init(jax.random.PRNGKey(0), dum_input)
    apply_fft = jax.jit(lambda x: ma_fft.apply(params_fft, x))

    params_matrix = ma_matrix.init(jax.random.PRNGKey(0), dum_input)
    apply_matrix = jax.jit(lambda x: ma_matrix.apply(params_matrix, x))

    print("timing DenseSymm fft mode, L = " + str(L))
    compile_time = timeit(lambda: apply_fft(dum_input).block_until_ready(), number=1)
    runtime = timeit(lambda: apply_fft(dum_input).block_until_ready(), number=10) / 10
    print("compile time = " + str(compile_time) + " runtime is " + str(runtime))

    print("timing DenseSymm matrix mode, L = " + str(L))
    compile_time = timeit(lambda: apply_matrix(dum_input).block_until_ready(), number=1)
    runtime = (
        timeit(lambda: apply_matrix(dum_input).block_until_ready(), number=10) / 10
    )
    print("compile time = " + str(compile_time) + " runtime is " + str(runtime))


# Test DenseEquivariant

for L in [2, 4, 6, 8, 10]:

    graph = Square(L)
    sg = graph.space_group()
    irreps = sg.irrep_matrices()
    product_table = sg.product_table

    ma_fft = DenseEquivariant(
        symmetries=product_table,
        mode="fft",
        shape=graph.extent,
        in_features=4,
        out_features=4,
    )
    ma_irreps = DenseEquivariant(
        symmetries=irreps, mode="irreps", in_features=4, out_features=4
    )
    ma_matrix = DenseEquivariant(
        symmetries=product_table, mode="matrix", in_features=4, out_features=4
    )

    dum_input = np.random.normal(0, 1, [1000, 4, len(sg)])

    params_fft = ma_fft.init(jax.random.PRNGKey(0), dum_input)
    apply_fft = jax.jit(lambda x: ma_fft.apply(params_fft, x))

    params_irreps = ma_irreps.init(jax.random.PRNGKey(0), dum_input)
    apply_irreps = jax.jit(lambda x: ma_irreps.apply(params_irreps, x))

    params_matrix = ma_matrix.init(jax.random.PRNGKey(0), dum_input)
    apply_matrix = jax.jit(lambda x: ma_matrix.apply(params_matrix, x))

    print("timing DenseEquivariant fft mode, L = " + str(L))
    compile_time = timeit(lambda: apply_fft(dum_input).block_until_ready(), number=1)
    runtime = timeit(lambda: apply_fft(dum_input).block_until_ready(), number=10) / 10
    print("compile time = " + str(compile_time) + " runtime is " + str(runtime))

    print("timing DenseEquivariant irreps mode, L = " + str(L))
    compile_time = timeit(lambda: apply_irreps(dum_input).block_until_ready(), number=1)
    runtime = (
        timeit(lambda: apply_irreps(dum_input).block_until_ready(), number=10) / 10
    )
    print("compile time = " + str(compile_time) + " runtime is " + str(runtime))

    print("timing DenseEquivariant matrix mode, L = " + str(L))
    compile_time = timeit(lambda: apply_matrix(dum_input).block_until_ready(), number=1)
    runtime = (
        timeit(lambda: apply_matrix(dum_input).block_until_ready(), number=10) / 10
    )
    print("compile time = " + str(compile_time) + " runtime is " + str(runtime))



from subprocess import check_output
import importlib


def exec_in_terminal(command):
    """Run a command in the terminal and get the
    output stripping the last newline.

    Args:
        command: a string or list of strings
    """
    return check_output(command).strip().decode("utf8")


def is_available(lib_name: str) -> bool:
    """
    Checks if a library can be imported
    """
    try:
        importlib.import_module(lib_name)
        available = True
    except ImportError:
        available = False

    return available


def version(lib_name) -> str:
    """
    Returns the version of a library as a string or
    unavailable if it cannot be imported
    """
    if is_available(lib_name):
        return _version(lib_name)
    else:
        return "unavailable"


def _version(lib_name):
    """
    Returns the version of a package.
    If version cannot be determined returns "available"
    """
    lib = importlib.import_module(lib_name)
    if hasattr(lib, "__version__"):
        return lib.__version__
    else:
        return "available"



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The score function (REINFORCE) gradient estimator of an expectation

from typing import Callable, Tuple
from functools import partial

import jax
from jax import numpy as jnp

from netket.stats import statistics as mpi_statistics, mean as mpi_mean, Stats
from netket.utils.types import PyTree

from ._vjp import vjp as nkvjp


def expect(
    log_pdf: Callable[[PyTree, jnp.ndarray], jnp.ndarray],
    expected_fun: Callable[[PyTree, jnp.ndarray], jnp.ndarray],
    pars: PyTree,
    σ: jnp.ndarray,
    *expected_fun_args,
    n_chains: int = None,
) -> Tuple[jnp.ndarray, Stats]:
    """
    Computes the expectation value over a log-pdf.

    Args:
        log_pdf:
        expected_ffun
    """
    return _expect(n_chains, log_pdf, expected_fun, pars, σ, *expected_fun_args)


# log_prob_args and integrand_args are independent of params when taking the
# gradient. They can be continuous or discrete, and they can be pytrees
# Does not support higher-order derivatives yet
@partial(jax.custom_vjp, nondiff_argnums=(0, 1, 2))
def _expect(n_chains, log_pdf, expected_fun, pars, σ, *expected_fun_args):
    L_σ = expected_fun(pars, σ, *expected_fun_args)
    if n_chains is not None:
        L_σ = L_σ.reshape((n_chains, -1))

    L̄_σ = mpi_statistics(L_σ.T)
    # L̄_σ = L_σ.mean(axis=0)

    return L̄_σ.mean, L̄_σ


def _expect_fwd(n_chains, log_pdf, expected_fun, pars, σ, *expected_fun_args):
    L_σ = expected_fun(pars, σ, *expected_fun_args)
    if n_chains is not None:
        L_σ_r = L_σ.reshape((n_chains, -1))
    else:
        L_σ_r = L_σ

    L̄_stat = mpi_statistics(L_σ_r.T)

    L̄_σ = L̄_stat.mean
    # L̄_σ = L_σ.mean(axis=0)

    # Use the baseline trick to reduce the variance
    ΔL_σ = L_σ - L̄_σ

    return (L̄_σ, L̄_stat), (pars, σ, expected_fun_args, ΔL_σ)


# TODO: in principle, the gradient of an expectation is another expectation,
# so it should support higher-order derivatives
# But I don't know how to transform log_prob_fun into grad(log_prob_fun) while
# keeping the batch dimension and without a loop through the batch dimension
def _expect_bwd(n_chains, log_pdf, expected_fun, residuals, dout):
    pars, σ, cost_args, ΔL_σ = residuals
    dL̄, dL̄_stats = dout

    def f(pars, σ, *cost_args):
        log_p = log_pdf(pars, σ)
        term1 = jax.vmap(jnp.multiply)(ΔL_σ, log_p)
        term2 = expected_fun(pars, σ, *cost_args)
        out = mpi_mean(term1 + term2, axis=0)
        out = out.sum()
        return out

    _, pb = nkvjp(f, pars, σ, *cost_args)
    grad_f = pb(dL̄)
    return grad_f


_expect.defvjp(_expect_fwd, _expect_bwd)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .rbm import RBM, RBMModPhase, RBMMultiVal, RBMSymm
from .equivariant import GCNN
from .jastrow import Jastrow
from .mps import MPSPeriodic
from .gaussian import Gaussian
from .ndm import NDM
from .autoreg import AbstractARNN, ARNNDense, ARNNConv1D, ARNNConv2D
from .fast_autoreg import FastARNNDense, FastARNNConv1D, FastARNNConv2D

from netket.utils import _hide_submodules

_hide_submodules(__name__)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netket as nk

# 1D Lattice
L = 20  # 10

g = nk.graph.Hypercube(length=L, n_dim=1, pbc=True)

# Hilbert space of spins on the graph
hi = nk.hilbert.Spin(s=1 / 2, N=g.n_nodes)

# Ising spin hamiltonian
ha = nk.operator.Ising(hilbert=hi, graph=g, h=1.0)

# RBM Spin Machine
ma = nk.models.RBM(alpha=1, use_visible_bias=True, dtype=float)

# Metropolis Local Sampling
sa = nk.sampler.MetropolisLocal(hi, n_chains=16, reset_chains=False)

# Optimizer
op = nk.optimizer.Sgd(learning_rate=0.1)

# Variational monte carlo driver
gs = nk.VMC(ha, op, sa, ma, n_samples=1000)

# Run the optimization for 300 iterations
gs.run(n_iter=300, out="test")



# Copyright 2020 The Netket Authors. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket import legacy as nk
import numpy as np

# 1D Lattice
L = 20
g = nk.graph.Hypercube(length=L, n_dim=1, pbc=True)

# Hilbert space of spins on the graph
hi = nk.hilbert.Spin(s=1 / 2, N=g.n_nodes)


# Defining the Ising hamiltonian (with sign problem here)
# Using local operators
sx = [[0, 1], [1, 0]]
sz = [[1, 0], [0, -1]]

ha = nk.operator.LocalOperator(hi)

for i in range(L):
    ha += nk.operator.LocalOperator(hi, sx, [i])
    ha += nk.operator.LocalOperator(hi, np.kron(sz, sz), [i, (i + 1) % L])


# RBM Spin Machine
ma = nk.machine.RbmSpinPhase(alpha=1, hilbert=hi)
ma.init_random_parameters(seed=1234, sigma=0.01)

# Metropolis Local Sampling
sa = nk.sampler.MetropolisLocal(machine=ma)

# Optimizer
op = nk.optimizer.AdaDelta(ma)

# Stochastic Reconfiguration
sr = nk.optimizer.SR(ma, diag_shift=0.1)

# Create the optimization driver
gs = nk.Vmc(hamiltonian=ha, sampler=sa, optimizer=op, n_samples=1000, sr=sr)

gs.run(out="test", n_iter=3000)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import pytest
from pytest import raises

import numpy as np
import jax
import netket as nk
import flax

from .. import common

pytestmark = common.skipif_mpi

nk.config.update("NETKET_EXPERIMENTAL", True)

SEED = 2148364

machines = {}

standard_init = flax.linen.initializers.normal()
NDM = partial(nk.models.NDM, bias_init=standard_init, visible_bias_init=standard_init)

machines["model:(R->C)"] = NDM(
    alpha=1,
    beta=1,
    dtype=float,
    kernel_init=nk.nn.initializers.normal(stddev=0.1),
    bias_init=nk.nn.initializers.normal(stddev=0.1),
)
# machines["model:(C->C)"] = RBM(
#    alpha=1,
#    dtype=complex,
#    kernel_init=nk.nn.initializers.normal(stddev=0.1),
#    bias_init=nk.nn.initializers.normal(stddev=0.1),
# )

operators = {}

L = 2
g = nk.graph.Hypercube(length=L, n_dim=1)
hi = nk.hilbert.Spin(s=0.5, N=L)

ha = nk.operator.Ising(hi, graph=g, h=1.0, dtype=complex)
jump_ops = [nk.operator.spin.sigmam(hi, i) for i in range(L)]

liouv = nk.operator.LocalLiouvillian(ha.to_local_operator(), jump_ops)

# operators["operator:(Lind)"] = liouv
operators["operator:(Lind^2)"] = liouv.H @ liouv
operators["operator:H"] = ha
operators["operator:sigmam"] = jump_ops[0]


@pytest.fixture(params=[pytest.param(ma, id=name) for name, ma in machines.items()])
def vstate(request):
    ma = request.param

    sa = nk.sampler.ExactSampler(hilbert=nk.hilbert.DoubledHilbert(hi), n_chains=16)

    vs = nk.vqs.MCMixedState(sa, ma, n_samples=1000, seed=SEED)

    return vs


def test_n_samples_api(vstate):
    with raises(
        ValueError,
    ):
        vstate.n_samples = -1

    with raises(
        ValueError,
    ):
        vstate.chain_length = -2

    with raises(
        ValueError,
    ):
        vstate.n_discard_per_chain = -1

    vstate.n_samples = 2
    assert vstate.samples.shape[0:2] == (1, vstate.sampler.n_chains)

    vstate.chain_length = 2
    assert vstate.n_samples == 2 * vstate.sampler.n_chains
    assert vstate.samples.shape[0:2] == (2, vstate.sampler.n_chains)

    vstate.n_samples = 1000
    vstate.n_discard_per_chain = None
    assert vstate.n_discard_per_chain == 0

    vstate.sampler = nk.sampler.MetropolisLocal(
        hilbert=nk.hilbert.DoubledHilbert(hi), n_chains=16
    )
    vstate.n_discard_per_chain = None
    assert vstate.n_discard_per_chain == vstate.n_samples // 10


def test_n_samples_diag_api(vstate):
    with raises(
        ValueError,
    ):
        vstate.n_samples_diag = -1

    with raises(
        ValueError,
    ):
        vstate.chain_length_diag = -2

    with raises(
        ValueError,
    ):
        vstate.n_discard_per_chain = -1

    vstate.n_samples_diag = 2
    assert (
        vstate.diagonal.samples.shape[0:2]
        == (1, vstate.sampler_diag.n_chains)
        == (1, vstate.diagonal.sampler.n_chains)
    )

    vstate.chain_length_diag = 2
    assert (
        vstate.n_samples_diag
        == 2 * vstate.sampler_diag.n_chains
        == 2 * vstate.diagonal.sampler.n_chains
    )
    assert (
        vstate.diagonal.samples.shape[0:2]
        == (2, vstate.diagonal.sampler.n_chains)
        == (2, vstate.sampler_diag.n_chains)
    )

    vstate.n_samples = 1000
    vstate.n_discard_per_chain = None
    assert vstate.n_discard_per_chain_diag == 0

    vstate.sampler_diag = nk.sampler.MetropolisLocal(hilbert=hi, n_chains=16)
    vstate.n_discard_per_chain_diag = None
    assert vstate.n_discard_per_chain_diag == vstate.n_samples_diag // 10


def test_deprecations(vstate):
    vstate.sampler_diag = nk.sampler.MetropolisLocal(hilbert=hi, n_chains=16)

    # deprecation
    with pytest.warns(FutureWarning):
        vstate.n_discard_diag = 10

    with pytest.warns(FutureWarning):
        vstate.n_discard_diag

    vstate.n_discard_diag = 10
    assert vstate.n_discard_diag == 10
    assert vstate.n_discard_per_chain_diag == 10


def test_serialization(vstate):
    from flax import serialization

    bdata = serialization.to_bytes(vstate)

    vstate_new = nk.vqs.MCMixedState(
        vstate.sampler, vstate.model, n_samples=10, seed=SEED + 313
    )

    vstate_new = serialization.from_bytes(vstate_new, bdata)

    jax.tree_multimap(
        np.testing.assert_allclose, vstate.parameters, vstate_new.parameters
    )
    np.testing.assert_allclose(vstate.samples, vstate_new.samples)
    np.testing.assert_allclose(vstate.diagonal.samples, vstate_new.diagonal.samples)
    assert vstate.n_samples == vstate_new.n_samples
    assert vstate.n_discard_per_chain == vstate_new.n_discard_per_chain
    assert vstate.n_samples_diag == vstate_new.n_samples_diag
    assert vstate.n_discard_per_chain_diag == vstate_new.n_discard_per_chain_diag


@pytest.mark.parametrize(
    "operator",
    [
        pytest.param(
            op,
            id=name,
        )
        for name, op in operators.items()
    ],
)
def test_expect_numpysampler_works(vstate, operator):
    sampl = nk.sampler.MetropolisLocalNumpy(vstate.hilbert)
    vstate.sampler = sampl
    out = vstate.expect(operator)
    assert isinstance(out, nk.stats.Stats)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

import jax
import jax.numpy as jnp

import netket as nk

from .. import common  # noqa: F401

WEIGHT_SEED = 3


@pytest.fixture()
def arr(request, _mpi_size):
    return jax.random.normal(jax.random.PRNGKey(WEIGHT_SEED), (100 * _mpi_size, 10))


@pytest.fixture()
def arr_loc(request, arr, _mpi_rank):
    return arr[100 * _mpi_rank : 100 * (_mpi_rank + 1), :]


@pytest.mark.parametrize("axis", [None, 0])
def test_mean(axis, arr, arr_loc):
    arr_mean = jnp.mean(arr, axis=axis)

    nk_mean = nk.stats.mean(arr_loc, axis=axis)
    np.testing.assert_allclose(nk_mean, arr_mean)

    nk_mean = nk.stats.mean(arr_loc, keepdims=True, axis=axis)
    assert nk_mean.ndim == arr.ndim
    np.testing.assert_allclose(nk_mean, arr_mean.reshape(1, -1))


@pytest.mark.parametrize("axis", [None, 0])
def test_subtract_mean(axis, arr, arr_loc, _mpi_rank):
    arr_sub = arr - jnp.mean(arr, axis=axis)

    nk_sub = nk.stats.subtract_mean(arr_loc, axis=axis)
    np.testing.assert_allclose(
        nk_sub, arr_sub[100 * _mpi_rank : 100 * (_mpi_rank + 1), :]
    )


@pytest.mark.parametrize("axis", [None, 0])
def test_sum(axis, arr, arr_loc):
    arr_sum = jnp.sum(arr, axis=axis)

    nk_sum = nk.stats.sum(arr_loc, axis=axis)
    np.testing.assert_allclose(nk_sum, arr_sum)

    nk_sum = nk.stats.sum(arr_loc, keepdims=True, axis=axis)
    assert nk_sum.ndim == arr.ndim
    np.testing.assert_allclose(nk_sum, arr_sum.reshape(1, -1))


@pytest.mark.parametrize("axis", [None, 0])
def test_var(axis, arr, arr_loc):
    arr_var = jnp.var(arr, axis=axis)

    nk_var = nk.stats.var(arr_loc, axis=axis)
    np.testing.assert_allclose(nk_var, arr_var)


@pytest.mark.parametrize("axis", [None, 0])
def test_total_size(axis, arr, arr_loc, _mpi_size):
    if axis is None:
        sz = arr.size
    else:
        sz = arr.shape[axis]
    sz = sz

    nk_sz = nk.stats.total_size(arr_loc, axis=axis)
    assert nk_sz == sz



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket.utils import warn_deprecation


# To be removed in v3.1
def graph_to_N_depwarn(N, graph):

    if graph is not None:
        warn_deprecation(
            r"""
            The ``graph`` argument for hilbert spaces has been deprecated in v3.0.
            It has been replaced by the argument ``N`` accepting an integer, with
            the number of nodes in the graph.

            You can update your code by passing `N=_your_graph.n_nodes`.
            If you are also using `Ising`, `Heisenberg`, `BoseHubbard` or `GraphOperator`
            Hamiltonians you must now provide them with the extra argument
            ``graph=_your_graph``, as they no longer take it from the Hilbert space.
            """
        )

        if N == 1:
            return graph.n_nodes
        else:
            raise ValueError(
                "Graph object can only take one argument among N and graph"
                "(deprecated)."
            )

    return N



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Tuple

import flax
from flax import linen as nn
from jax import lax
from jax import numpy as jnp
from netket.nn.initializers import lecun_normal, zeros
from netket.utils.types import Array, DType, NNInitFunc

default_kernel_init = lecun_normal()


def wrap_kernel_init(kernel_init, mask):
    """Correction to LeCun normal init."""

    corr = jnp.sqrt(mask.size / mask.sum())

    def wrapped_kernel_init(*args):
        return corr * mask * kernel_init(*args)

    return wrapped_kernel_init


class MaskedDense1D(nn.Module):
    """1D linear transformation module with mask for autoregressive NN."""

    features: int
    """number of output features, should be the last dimension."""
    exclusive: bool
    """True if an output element does not depend on the input element at the same index."""
    use_bias: bool = True
    """whether to add a bias to the output (default: True)."""
    dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""
    precision: Any = None
    """numerical precision of the computation, see `jax.lax.Precision` for details."""
    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the weight matrix."""
    bias_init: NNInitFunc = zeros
    """initializer for the bias."""

    @nn.compact
    def __call__(self, inputs: Array) -> Array:
        """
        Applies a masked linear transformation to the inputs.

        Args:
          inputs: input data with dimensions (batch, length, features).

        Returns:
          The transformed data.
        """
        dtype = jnp.promote_types(inputs.dtype, self.dtype)

        inputs = jnp.asarray(inputs, dtype)

        is_single_input = False
        if inputs.ndim == 2:
            is_single_input = True
            inputs = jnp.expand_dims(inputs, axis=0)

        batch, size, in_features = inputs.shape
        inputs = inputs.reshape((batch, size * in_features))

        mask = jnp.ones((size, size), dtype=self.dtype)
        mask = jnp.triu(mask, self.exclusive)
        mask = jnp.kron(mask, jnp.ones((in_features, self.features), dtype=self.dtype))

        kernel = self.param(
            "kernel",
            wrap_kernel_init(self.kernel_init, mask),
            (size * in_features, size * self.features),
            self.dtype,
        )
        mask = jnp.asarray(mask, dtype)
        kernel = jnp.asarray(kernel, dtype)

        y = lax.dot(inputs, mask * kernel, precision=self.precision)

        y = y.reshape((batch, size, self.features))

        if is_single_input:
            y = y.squeeze(axis=0)

        if self.use_bias:
            bias = self.param("bias", self.bias_init, (size, self.features), self.dtype)
            bias = jnp.asarray(bias, dtype)
            y = y + bias

        return y


class MaskedConv1D(nn.Module):
    """1D convolution module with mask for autoregressive NN."""

    features: int
    """number of convolution filters."""
    kernel_size: int
    """length of the convolutional kernel."""
    kernel_dilation: int
    """dilation factor of the convolution kernel."""
    exclusive: bool
    """True if an output element does not depend on the input element at the same index."""
    feature_group_count: int = 1
    """if specified, divides the input features into groups (default: 1)."""
    use_bias: bool = True
    """whether to add a bias to the output (default: True)."""
    dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""
    precision: Any = None
    """numerical precision of the computation, see `jax.lax.Precision` for details."""
    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the convolutional kernel."""
    bias_init: NNInitFunc = zeros
    """initializer for the bias."""

    @nn.compact
    def __call__(self, inputs: Array) -> Array:
        """
        Applies a masked convolution to the inputs.
        For 1D convolution, there is not really a mask. We only need to apply
        appropriate padding.

        Args:
          inputs: input data with dimensions (batch, length, features).

        Returns:
          The convolved data.
        """
        dtype = jnp.promote_types(inputs.dtype, self.dtype)

        inputs = jnp.asarray(inputs, dtype)

        kernel_size = self.kernel_size - self.exclusive
        dilation = self.kernel_dilation

        is_single_input = False
        if inputs.ndim == 2:
            is_single_input = True
            inputs = jnp.expand_dims(inputs, axis=0)

        in_features = inputs.shape[-1]
        assert in_features % self.feature_group_count == 0
        kernel_shape = (
            kernel_size,
            in_features // self.feature_group_count,
            self.features,
        )

        kernel = self.param("kernel", self.kernel_init, kernel_shape, self.dtype)
        kernel = jnp.asarray(kernel, dtype)

        if self.exclusive:
            inputs = inputs[:, :-dilation, :]

        # Zero padding
        y = jnp.pad(
            inputs,
            (
                (0, 0),
                ((kernel_size - (not self.exclusive)) * dilation, 0),
                (0, 0),
            ),
        )

        dimension_numbers = flax.linen.linear._conv_dimension_numbers(inputs.shape)
        y = lax.conv_general_dilated(
            y,
            kernel,
            window_strides=(1,),
            padding="VALID",
            lhs_dilation=(1,),
            rhs_dilation=(dilation,),
            dimension_numbers=dimension_numbers,
            feature_group_count=self.feature_group_count,
            precision=self.precision,
        )

        if is_single_input:
            y = y.squeeze(axis=0)

        if self.use_bias:
            bias = self.param("bias", self.bias_init, (self.features,), self.dtype)
            bias = jnp.asarray(bias, dtype)
            y = y + bias

        return y


class MaskedConv2D(nn.Module):
    """2D convolution module with mask for autoregressive NN."""

    features: int
    """number of convolution filters."""
    kernel_size: Tuple[int, int]
    """shape of the convolutional kernel `(h, w)`. Typically, `h = w // 2 + 1`."""
    kernel_dilation: Tuple[int, int]
    """a sequence of 2 integers, giving the dilation factor to
    apply in each spatial dimension of the convolution kernel."""
    exclusive: bool
    """True if an output element does not depend on the input element at the same index."""
    feature_group_count: int = 1
    """if specified, divides the input features into groups (default: 1)."""
    use_bias: bool = True
    """whether to add a bias to the output (default: True)."""
    dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""
    precision: Any = None
    """numerical precision of the computation, see `jax.lax.Precision` for details."""
    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the convolutional kernel."""
    bias_init: NNInitFunc = zeros
    """initializer for the bias."""

    def setup(self):
        kernel_h, kernel_w = self.kernel_size
        mask = jnp.ones((kernel_h, kernel_w, 1, 1), dtype=self.dtype)
        mask = mask.at[-1, kernel_w // 2 + (not self.exclusive) :].set(0)
        self.mask = mask

    @nn.compact
    def __call__(self, inputs: Array) -> Array:
        """
        Applies a masked convolution to the inputs.

        Args:
          inputs: input data with dimensions (batch, width, height, features).

        Returns:
          The convolved data.
        """
        dtype = jnp.promote_types(inputs.dtype, self.dtype)

        inputs = jnp.asarray(inputs, dtype)

        kernel_h, kernel_w = self.kernel_size
        dilation_h, dilation_w = self.kernel_dilation
        ones = (1, 1)

        is_single_input = False
        if inputs.ndim == 3:
            is_single_input = True
            inputs = jnp.expand_dims(inputs, axis=0)

        in_features = inputs.shape[-1]
        assert in_features % self.feature_group_count == 0
        kernel_shape = self.kernel_size + (
            in_features // self.feature_group_count,
            self.features,
        )

        kernel = self.param(
            "kernel",
            wrap_kernel_init(self.kernel_init, self.mask),
            kernel_shape,
            self.dtype,
        )
        mask = jnp.asarray(self.mask, dtype)
        kernel = jnp.asarray(kernel, dtype)

        # Zero padding
        y = jnp.pad(
            inputs,
            (
                (0, 0),
                ((kernel_h - 1) * dilation_h, 0),
                (kernel_w // 2 * dilation_w, (kernel_w - 1) // 2 * dilation_w),
                (0, 0),
            ),
        )

        dimension_numbers = flax.linen.linear._conv_dimension_numbers(inputs.shape)
        y = lax.conv_general_dilated(
            y,
            mask * kernel,
            window_strides=ones,
            padding="VALID",
            lhs_dilation=ones,
            rhs_dilation=self.kernel_dilation,
            dimension_numbers=dimension_numbers,
            feature_group_count=self.feature_group_count,
            precision=self.precision,
        )

        if is_single_input:
            y = y.squeeze(axis=0)

        if self.use_bias:
            bias = self.param("bias", self.bias_init, (self.features,), self.dtype)
            bias = jnp.asarray(bias, dtype)
            y = y + bias

        return y



# Copyright 2020, 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Union

import numpy as np


@dataclass
class EarlyStopping:
    """A simple callback to stop NetKet if there are no more improvements in the training.
    based on `driver._loss_name`."""

    min_delta: float = 0.0
    """Minimum change in the monitored quantity to qualify as an improvement."""
    patience: Union[int, float] = 0
    """Number of epochs with no improvement after which training will be stopped."""
    baseline: float = None
    """Baseline value for the monitored quantity. Training will stop if the driver hits the baseline."""
    monitor: str = "mean"
    """Loss statistic to monitor. Should be one of 'mean', 'variance', 'sigma'."""

    def __post_init__(self):
        self._best_val = np.infty
        self._best_iter = 0

    def __call__(self, step, log_data, driver):
        """
        A boolean function that determines whether or not to stop training.

        Args:
            step: An integer corresponding to the step (iteration or epoch) in training.
            log_data: A dictionary containing log data for training.
            driver: A NetKet variational driver.

        Returns:
            A boolean. If True, training continues, else, it does not.
        """
        loss = np.real(getattr(log_data[driver._loss_name], self.monitor))
        if loss <= self._best_val:
            self._best_val = loss
            self._best_iter = step
        if self.baseline is not None:
            if loss <= self.baseline:
                return False
        if (
            step - self._best_iter >= self.patience
            and loss > self._best_val - self.min_delta
        ):
            return False
        else:
            return True



from functools import partial
from io import StringIO

import pytest
from pytest import approx, raises

import numpy as np
import jax
import jax.numpy as jnp
import netket as nk

from contextlib import redirect_stderr
import tempfile
import re

from .. import common

pytestmark = common.skipif_mpi

SEED = 214748364


def _setup_vmc(dtype=np.float32, sr=True):
    L = 4
    g = nk.graph.Hypercube(length=L, n_dim=1)
    hi = nk.hilbert.Spin(s=0.5, N=g.n_nodes)

    ma = nk.models.RBM(alpha=1, dtype=dtype)
    sa = nk.sampler.ExactSampler(hilbert=hi, n_chains=16)

    vs = nk.vqs.MCState(sa, ma, n_samples=1000, seed=SEED)

    ha = nk.operator.Ising(hi, graph=g, h=1.0)
    op = nk.optimizer.Sgd(learning_rate=0.05)

    # Add custom observable
    X = [[0, 1], [1, 0]]
    sx = nk.operator.LocalOperator(hi, [X] * L, [[i] for i in range(8)])

    if sr:
        sr_config = nk.optimizer.SR()
    else:
        sr_config = None
    driver = nk.VMC(ha, op, variational_state=vs, preconditioner=sr_config)

    return ha, sx, vs, sa, driver


#####
@partial(jax.jit, static_argnums=0)
@partial(jax.vmap, in_axes=(None, None, 0, 0, 0), out_axes=(0))
def local_value_kernel(logpsi, pars, σ, σp, mel):
    return jnp.sum(mel * jnp.exp(logpsi(pars, σp) - logpsi(pars, σ)))


def local_values(logpsi, variables, Ô, σ):
    """
    Returns loc_vals for an operator O
    """
    σp, mels = Ô.get_conn_padded(np.asarray(σ).reshape((-1, σ.shape[-1])))
    loc_vals = local_value_kernel(
        logpsi, variables, σ.reshape((-1, σ.shape[-1])), σp, mels
    )
    return loc_vals.reshape(σ.shape[:-1])


def estimate_gradient(logpsi, variables, Ô, σ):
    """
    Estimate the gradient of hermmitian operator O using legacy code
    """
    O_loc = local_values(logpsi, variables, Ô, σ)
    O_loc -= O_loc.mean()
    O_loc /= np.prod(O_loc.shape)

    σr = σ.reshape((-1, σ.shape[-1]))
    O_loc_r = O_loc.reshape(-1)

    _, vjp = nk.jax.vjp(logpsi, variables, σr, conjugate=True)
    Opsi = vjp(O_loc_r.conjugate())[0]

    return Opsi


####


def test_estimate():
    ha, *_, driver = _setup_vmc()
    driver.estimate(ha)
    driver.advance(1)
    driver.estimate(ha)


def test_raise_n_iter():
    ha, sx, ma, sampler, driver = _setup_vmc()
    with raises(
        ValueError,
    ):
        driver.run("prova", 12)


def test_reset():
    ha, *_, driver = _setup_vmc()
    assert driver.step_count == 0
    driver.advance(1)
    assert driver.step_count == 1
    driver.reset()
    assert driver.step_count == 0


def test_vmc_construction_vstate():
    ha, sx, ma, sa, driver = _setup_vmc()

    op = nk.optimizer.Sgd(learning_rate=0.05)

    driver = nk.Vmc(ha, op, sa, nk.models.RBM(), n_samples=1000, seed=SEED)

    driver.run(1)

    assert driver.step_count == 1

    with raises(TypeError):
        ha2 = nk.operator.LocalOperator(ha.hilbert * ha.hilbert)
        driver = nk.VMC(ha2, op, variational_state=driver.state)


def test_vmc_functions():
    ha, sx, ma, sampler, driver = _setup_vmc()

    driver.advance(500)

    assert driver.energy.mean == approx(ma.expect(ha).mean, abs=1e-5)

    state = ma.to_array()

    n_samples = 16000
    ma.n_samples = n_samples
    ma.n_discard_per_chain = 100

    # Check zero gradient
    _, grads = ma.expect_and_grad(ha)

    def check_shape(a, b):
        assert a.shape == b.shape

    jax.tree_multimap(check_shape, grads, ma.parameters)
    grads, _ = nk.jax.tree_ravel(grads)

    assert np.mean(np.abs(grads) ** 2) == approx(0.0, abs=1e-8)
    # end

    for op, name in (ha, "ha"), (sx, "sx"):
        print("Testing expectation of op={}".format(name))

        exact_ex = (state.T.conj() @ op.to_linear_operator() @ state).real

        op_stats = ma.expect(op)

        mean = op_stats.mean
        var = op_stats.variance
        print(mean, var)

        # 5-sigma test for expectation values
        tol = np.sqrt(var / float(ma.n_samples)) * 5
        assert mean.real == approx(exact_ex, abs=tol)


def test_vmc_progress_bar():
    ha, sx, ma, sampler, driver = _setup_vmc()
    tempdir = tempfile.mkdtemp()
    prefix = tempdir + "/vmc_progressbar_test"

    f = StringIO()
    with redirect_stderr(f):
        driver.run(5, prefix)
    pbar = f.getvalue().split("\r")[-1]
    assert re.match(r"100%\|#*\| (\d+)/\1", pbar)
    assert re.search(r"Energy=[-+]?[0-9]*\.?[0-9]*", pbar)
    assert re.search(r"σ²=[-+]?[0-9]*\.?[0-9]*", pbar)

    f = StringIO()
    with redirect_stderr(f):
        driver.run(5, prefix, show_progress=None)
    pbar = f.getvalue()
    assert not len(pbar)


def _energy(par, vstate, H):
    vstate.parameters = par
    psi = vstate.to_array()
    return np.real(psi.conj() @ H @ psi)


def central_diff_grad(func, x, eps, *args):
    grad = np.zeros(len(x), dtype=x.dtype)
    epsd = np.zeros(len(x), dtype=x.dtype)
    epsd[0] = eps
    for i in range(len(x)):
        assert not np.any(np.isnan(x + epsd))
        grad_r = 0.5 * (func(x + epsd, *args) - func(x - epsd, *args))
        if nk.jax.is_complex(x):
            grad_i = 0.5 * (func(x + 1j * epsd, *args) - func(x - 1j * epsd, *args))
            grad[i] = 0.5 * grad_r + 0.5j * grad_i
        else:
            grad_i = 0.0
            grad[i] = 0.5 * grad_r

        assert not np.isnan(grad[i])
        grad[i] /= eps
        epsd = np.roll(epsd, 1)
    return grad


def same_derivatives(der_log, num_der_log, abs_eps=1.0e-6, rel_eps=1.0e-6):
    assert der_log.shape == num_der_log.shape

    np.testing.assert_allclose(
        der_log.real, num_der_log.real, rtol=rel_eps, atol=abs_eps
    )

    # The imaginary part is a bit more tricky, there might be an arbitrary phase shift
    assert np.max(
        np.abs(np.exp(np.imag(der_log - num_der_log) * 1.0j) - 1.0)
    ) == approx(0.0, rel=rel_eps, abs=abs_eps)


@pytest.mark.parametrize("dtype", [np.float64, np.complex128])
def test_vmc_gradient(dtype):
    ha, sx, ma, sampler, driver = _setup_vmc(dtype=dtype, sr=False)
    driver.run(3, out=None)

    # old formula
    _, grads = ma.expect_and_grad(ha)
    grad_old = estimate_gradient(
        lambda w, x: ma._apply_fun({"params": w}, x), ma.parameters, ha, ma.samples
    )

    # check that the old fformula and the new give same result
    jax.tree_multimap(lambda x, y: same_derivatives(x, y), grads, grad_old)
    # end old

    pars_0 = ma.parameters
    pars, unravel = nk.jax.tree_ravel(pars_0)

    def energy_fun(par, vstate, H):
        return _energy(unravel(par), vstate, H)

    grad_exact = central_diff_grad(energy_fun, pars, 1.0e-5, ma, ha.to_sparse())

    driver.state.n_samples = 1e5
    driver.state.n_discard_per_chain = 1e3
    driver.state.parameters = pars_0
    _, _grad_approx = ma.expect_and_grad(ha)  # driver._forward_and_backward()
    grad_approx, _ = nk.jax.tree_ravel(_grad_approx)

    err = 6 / np.sqrt(driver.state.n_samples)  # improve error bound
    same_derivatives(grad_approx, grad_exact, abs_eps=err, rel_eps=1.0e-3)


def test_vmc_sr_legacy_api():
    ha, sx, ma, sampler, driver = _setup_vmc(sr=True)
    op = driver.optimizer
    vs = driver.state
    sr_config = driver.preconditioner

    with pytest.warns(FutureWarning):
        driver = nk.VMC(
            ha,
            op,
            variational_state=vs,
            sr=sr_config,
        )

    with pytest.warns(FutureWarning):
        driver = nk.VMC(
            ha,
            op,
            variational_state=vs,
            preconditioner=sr_config,
            sr_restart=True,
        )

    with pytest.raises(ValueError):
        driver = nk.VMC(
            ha, op, variational_state=vs, sr=sr_config, preconditioner=sr_config
        )



# Copyright 2020 The Netket Authors. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket import legacy as nk
import numpy as np

np.set_printoptions(linewidth=180)
rg = nk.utils.RandomEngine(seed=1234)

# 1D Lattice
L = 9
g = nk.graph.Hypercube(length=L, n_dim=1, pbc=False)

# Hilbert space of spins on the graph
hi = nk.hilbert.Spin(s=1 / 2, N=g.n_nodes)


# Defining the Ising hamiltonian (with sign problem here)
# Using local operators
sx = [[0, 1], [1, 0]]
sy = [[0, -1j], [1j, 0]]
sz = [[1, 0], [0, -1]]
s0 = [[0, 0], [0, 0]]

sigmam = [[0, 0], [1, 0]]

ha = nk.operator.LocalOperator(hi)
j_ops = []

for i in range(L):
    ha += nk.operator.LocalOperator(hi, sx, [i])
    ha += nk.operator.LocalOperator(hi, np.kron(sz, sz), [i, (i + 1) % L])
    j_ops.append(nk.operator.LocalOperator(hi, sigmam, [i]))


#  Create the lindbladian with no jump operators
lind = nk.operator.LocalLiouvillian(ha)

# add the jump operators
for j_op in j_ops:
    lind.add_jump_op(j_op)


rho = nk.exact.steady_state(
    lind, method="iterative", sparse=True, maxiter=1000, tol=1e-5
)



from functools import partial
from typing import Callable, Union

import numpy as np

import jax
from jax import numpy as jnp

from flax import linen as nn

from netket import jax as nkjax
from netket.stats import Stats
from netket.utils.types import PyTree
from netket.utils.dispatch import dispatch

from netket.operator import (
    AbstractOperator,
    AbstractSuperOperator,
    Squared,
)

from .mc_state import MCState
from .mc_mixed_state import MCMixedState

AFunType = Callable[[nn.Module, PyTree, jnp.ndarray], jnp.ndarray]
ATrainFunType = Callable[
    [nn.Module, PyTree, jnp.ndarray, Union[bool, PyTree]], jnp.ndarray
]


def _check_hilbert(A, B):
    if A.hilbert != B.hilbert:
        raise NotImplementedError(  # pragma: no cover
            f"Non matching hilbert spaces {A.hilbert} and {B.hilbert}"
        )


def local_value_kernel(logpsi, pars, σ, σp, mel):
    """
    local_value kernel for MCState and generic operators
    """
    return jnp.sum(mel * jnp.exp(logpsi(pars, σp) - logpsi(pars, σ)))


def local_value_squared_kernel(logpsi, pars, σ, σp, mel):
    """
    local_value kernel for MCState and Squared (generic) operators
    """
    return jnp.abs(local_value_kernel(logpsi, pars, σ, σp, mel)) ** 2


def local_value_op_op_cost(logpsi, pars, σ, σp, mel):
    """
    local_value kernel for MCMixedState and generic operators
    """
    σ_σp = jax.vmap(lambda σp, σ: jnp.hstack((σp, σ)), in_axes=(0, None))(σp, σ)
    σ_σ = jnp.hstack((σ, σ))
    return jnp.sum(mel * jnp.exp(logpsi(pars, σ_σp) - logpsi(pars, σ_σ)))


@dispatch.multi((MCState, Squared), (MCMixedState, Squared))
def expect(vstate: MCState, Ô: Squared) -> Stats:  # noqa: F811
    _check_hilbert(vstate, Ô)

    σ = vstate.samples

    σp, mels = Ô.parent.get_conn_padded(np.asarray(σ).reshape((-1, σ.shape[-1])))

    return _expect(
        vstate.sampler.machine_pow,
        vstate._apply_fun,
        local_value_squared_kernel,
        vstate.parameters,
        vstate.model_state,
        σ,
        σp,
        mels,
    )


@dispatch.multi((MCState, AbstractOperator), (MCMixedState, AbstractSuperOperator))
def expect(vstate: MCState, Ô: AbstractOperator) -> Stats:  # noqa: F811
    _check_hilbert(vstate, Ô)

    σ = vstate.samples

    σp, mels = Ô.get_conn_padded(np.asarray(σ).reshape((-1, σ.shape[-1])))

    return _expect(
        vstate.sampler.machine_pow,
        vstate._apply_fun,
        local_value_kernel,
        vstate.parameters,
        vstate.model_state,
        σ,
        σp,
        mels,
    )


@dispatch
def expect(vstate: MCMixedState, Ô: AbstractOperator) -> Stats:  # noqa: F811
    _check_hilbert(vstate.diagonal, Ô)

    σ = vstate.diagonal.samples

    σp, mels = Ô.get_conn_padded(np.asarray(σ).reshape((-1, σ.shape[-1])))

    return _expect(
        vstate.sampler.machine_pow,
        vstate._apply_fun,
        local_value_op_op_cost,
        vstate.parameters,
        vstate.model_state,
        σ,
        σp,
        mels,
    )


@partial(jax.jit, static_argnums=(1, 2))
def _expect(
    machine_pow: int,
    model_apply_fun: Callable,
    local_value_kernel: Callable,
    parameters: PyTree,
    model_state: PyTree,
    σ: jnp.ndarray,
    σp: jnp.ndarray,
    mels: jnp.ndarray,
) -> Stats:
    σ_shape = σ.shape

    if jnp.ndim(σ) != 2:
        σ = σ.reshape((-1, σ_shape[-1]))

    def logpsi(w, σ):
        return model_apply_fun({"params": w, **model_state}, σ)

    def log_pdf(w, σ):
        return machine_pow * model_apply_fun({"params": w, **model_state}, σ).real

    local_value_vmap = jax.vmap(
        partial(local_value_kernel, logpsi),
        in_axes=(None, 0, 0, 0),
        out_axes=0,
    )

    _, Ō_stats = nkjax.expect(
        log_pdf, local_value_vmap, parameters, σ, σp, mels, n_chains=σ_shape[0]
    )

    return Ō_stats



import netket as nk
import numpy as np

import pytest
import jax

operators = {}

# Ising 1D
g = nk.graph.Hypercube(length=20, n_dim=1, pbc=True)
hi = nk.hilbert.Spin(s=0.5, N=g.n_nodes)
operators["Ising 1D"] = nk.operator.Ising(hi, g, h=1.321)

# Heisenberg 1D
g = nk.graph.Hypercube(length=20, n_dim=1, pbc=True)
hi = nk.hilbert.Spin(s=0.5, total_sz=0, N=g.n_nodes)
operators["Heisenberg 1D"] = nk.operator.Heisenberg(hilbert=hi, graph=g)

# Bose Hubbard
g = nk.graph.Hypercube(length=3, n_dim=2, pbc=True)
hi = nk.hilbert.Fock(n_max=3, n_particles=6, N=g.n_nodes)
operators["Bose Hubbard"] = nk.operator.BoseHubbard(U=4.0, hilbert=hi, graph=g)

# Graph Hamiltonian
N = 20
sigmax = np.asarray([[0, 1], [1, 0]])
mszsz = np.asarray([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
edges = [[i, i + 1] for i in range(N - 1)] + [[N - 1, 0]]

g = nk.graph.Graph(edges=edges)
hi = nk.hilbert.CustomHilbert(local_states=[-1, 1], N=g.n_nodes)
operators["Graph Hamiltonian"] = nk.operator.GraphOperator(
    hi, g, site_ops=[sigmax], bond_ops=[mszsz]
)

# Graph Hamiltonian with colored edges
edges_c = [(i, j, i % 2) for i, j in edges]
g = nk.graph.Graph(edges=edges_c)
hi = nk.hilbert.CustomHilbert(local_states=[-1, 1], N=g.n_nodes)
operators["Graph Hamiltonian (colored edges)"] = nk.operator.GraphOperator(
    hi,
    g,
    site_ops=[sigmax],
    bond_ops=[1.0 * mszsz, 2.0 * mszsz],
    bond_ops_colors=[0, 1],
)

# Custom Hamiltonian
sx = [[0, 1], [1, 0]]
sy = [[0, -1.0j], [1.0j, 0]]
sz = [[1, 0], [0, -1]]
g = nk.graph.Graph(edges=[[i, i + 1] for i in range(20)])
hi = nk.hilbert.CustomHilbert(local_states=[-1, 1], N=g.n_nodes)


def _loc(*args):
    return nk.operator.LocalOperator(hi, *args)


sx_hat = _loc([sx] * 3, [[0], [1], [5]])
sy_hat = _loc([sy] * 4, [[2], [3], [4], [9]])
szsz_hat = _loc(sz, [0]) @ _loc(sz, [1])
szsz_hat += _loc(sz, [4]) @ _loc(sz, [5])
szsz_hat += _loc(sz, [6]) @ _loc(sz, [8])
szsz_hat += _loc(sz, [7]) @ _loc(sz, [0])

operators["Custom Hamiltonian"] = sx_hat + sy_hat + szsz_hat
operators["Custom Hamiltonian Prod"] = sx_hat * 1.5 + (2.0 * sy_hat)

operators["Pauli Hamiltonian (XX)"] = nk.operator.PauliStrings(["XX"], [0.1])
operators["Pauli Hamiltonian (XX+YZ+IZ)"] = nk.operator.PauliStrings(
    ["XX", "YZ", "IZ"], [0.1, 0.2, -1.4]
)


op_special = {}
for name, op in operators.items():
    if hasattr(op, "to_local_operator"):
        op_special[name] = op


@pytest.mark.parametrize(
    "op", [pytest.param(op, id=name) for name, op in operators.items()]
)
def test_produce_elements_in_hilbert(op):
    rng = nk.jax.PRNGSeq(0)
    hi = op.hilbert
    assert len(hi.local_states) == hi.local_size
    assert hi.size > 0

    local_states = hi.local_states

    max_conn_size = op.max_conn_size

    for i in range(1000):
        rstate = hi.random_state(rng.next())

        rstatet, mels = op.get_conn(rstate)

        assert np.all(np.isin(rstatet, local_states))
        assert len(mels) <= max_conn_size


@pytest.mark.parametrize(
    "op", [pytest.param(op, id=name) for name, op in operators.items()]
)
def test_is_hermitean(op):
    rng = nk.jax.PRNGSeq(0)

    hi = op.hilbert
    assert len(hi.local_states) == hi.local_size

    rstate = np.zeros(hi.size)

    for i in range(100):
        rstate = hi.random_state(rng.next())
        rstatet, mels = op.get_conn(rstate)

        for k, state in enumerate(rstatet):

            invstates, mels1 = op.get_conn(state)

            found = False
            for kp, invstate in enumerate(invstates):
                if np.array_equal(rstate, invstate.flatten()):
                    found = True
                    assert mels1[kp] == np.conj(mels[k])
                    break

            assert found


@pytest.mark.parametrize(
    "op", [pytest.param(op, id=name) for name, op in operators.items()]
)
def test_repr(op):
    assert type(op).__name__ in repr(op)


@pytest.mark.parametrize(
    "op", [pytest.param(op, id=name) for name, op in operators.items()]
)
def test_get_conn_numpy_closure(op):
    hi = op.hilbert
    closure = op._get_conn_flattened_closure()
    v = hi.random_state(jax.random.PRNGKey(0), 120)
    conn = np.empty(v.shape[0], dtype=np.intp)

    vp, mels = closure(np.asarray(v), conn)
    vp2, mels2 = op.get_conn_flattened(v, conn, pad=False)

    np.testing.assert_equal(vp, vp2)
    np.testing.assert_equal(mels, mels2)


@pytest.mark.parametrize(
    "op", [pytest.param(op, id=name) for name, op in operators.items()]
)
@pytest.mark.parametrize(
    "shape",
    [
        pytest.param(s, id=f"shape={s}")
        for s in [
            (2,),
            (
                2,
                1,
            ),
            (2, 1, 1),
        ]
    ],
)
def test_get_conn_padded(op, shape):
    hi = op.hilbert

    v = hi.random_state(jax.random.PRNGKey(0), shape)

    vp, mels = op.get_conn_padded(v)

    assert vp.ndim == v.ndim + 1
    assert mels.ndim == v.ndim
    print(mels.shape)
    print(vp.shape)

    vp_f, mels_f = op.get_conn_padded(v.reshape(-1, hi.size))
    np.testing.assert_allclose(vp_f, vp.reshape(-1, *vp.shape[-2:]))
    np.testing.assert_allclose(mels_f, mels.reshape(-1, mels.shape[-1]))


@pytest.mark.parametrize(
    "op", [pytest.param(op, id=name) for name, op in op_special.items()]
)
def test_to_local_operator(op):
    op.to_local_operator()
    # TODO check dense representaiton.


def test_no_segfault():
    g = nk.graph.Hypercube(8, 1)
    hi = nk.hilbert.Spin(0.5, N=g.n_nodes)

    lo = nk.operator.LocalOperator(hi, [[1, 0], [0, 1]], [0])
    lo = lo.transpose()

    hi = None

    lo = lo @ lo

    assert True


def test_deduced_hilbert_pauli():
    op = nk.operator.PauliStrings(["XXI", "YZX", "IZX"], [0.1, 0.2, -1.4])
    assert op.hilbert.size == 3
    assert len(op.hilbert.local_states) == 2
    assert np.allclose(op.hilbert.local_states, (0, 1))


def test_Heisenberg():
    g = nk.graph.Hypercube(8, 1)
    hi = nk.hilbert.Spin(0.5) ** 8

    def gs_energy(ham):
        return nk.exact.lanczos_ed(ham)

    ha1 = nk.operator.Heisenberg(hi, graph=g)
    ha2 = nk.operator.Heisenberg(hi, graph=g, J=2.0)

    assert 2 * gs_energy(ha1) == pytest.approx(gs_energy(ha2))

    ha1 = nk.operator.Heisenberg(hi, graph=g, sign_rule=True)
    ha2 = nk.operator.Heisenberg(hi, graph=g, sign_rule=False)

    assert gs_energy(ha1) == pytest.approx(gs_energy(ha2))

    with pytest.raises(
        ValueError, match=r"sign_rule=True specified for a non-bipartite lattice"
    ):
        g = nk.graph.Hypercube(7, 1)
        hi = nk.hilbert.Spin(0.5, N=g.n_nodes)

        assert not g.is_bipartite()

        nk.operator.Heisenberg(hi, graph=g, sign_rule=True)


def test_pauli():
    op = nk.operator.PauliStrings(["XX", "YZ", "IZ"], [0.1, 0.2, -1.4])

    op_l = (
        0.1
        * nk.operator.spin.sigmax(op.hilbert, 0, dtype=complex)
        * nk.operator.spin.sigmax(op.hilbert, 1)
    )
    op_l += (
        0.2
        * nk.operator.spin.sigmay(op.hilbert, 0)
        * nk.operator.spin.sigmaz(op.hilbert, 1)
    )
    op_l -= 1.4 * nk.operator.spin.sigmaz(op.hilbert, 1)

    assert np.allclose(op.to_dense(), op_l.to_dense())

    assert op.to_sparse().shape == op_l.to_sparse().shape


def test_pauli_order():
    """Check related to PR #836"""
    coeff1 = 1 + 0.9j
    coeff2 = 0.3 + 0.43j
    op = nk.operator.PauliStrings(["IZXY", "ZZYX"], [coeff1, coeff2])
    op1 = nk.operator.PauliStrings(["IZXY"], [coeff1])
    op2 = nk.operator.PauliStrings(["ZZYX"], [coeff2])
    op1_true = (
        coeff1
        * nk.operator.spin.sigmaz(op.hilbert, 1, dtype=complex)
        * nk.operator.spin.sigmax(op.hilbert, 2)
        * nk.operator.spin.sigmay(op.hilbert, 3)
    )
    op2_true = (
        coeff2
        * nk.operator.spin.sigmaz(op.hilbert, 0, dtype=complex)
        * nk.operator.spin.sigmaz(op.hilbert, 1)
        * nk.operator.spin.sigmay(op.hilbert, 2)
        * nk.operator.spin.sigmax(op.hilbert, 3)
    )
    assert np.allclose(op1.to_dense(), op1_true.to_dense())
    assert np.allclose(op2.to_dense(), op2_true.to_dense())
    assert np.allclose(op.to_dense(), (op1_true.to_dense() + op2_true.to_dense()))

    v = op.hilbert.all_states()
    vp, mels = op.get_conn_padded(v)
    assert vp.shape[1] == 1
    assert mels.shape[1] == 1



# Copyright 2020, 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Deprecation warnings for netket.optim, renamed to netket.optimizer

from netket.utils import warn_deprecation as _warn_deprecation

from . import optimizer as _new_module

_old_module_name = "netket.optim"
_new_module_name = "netket.optimizer"

_deprecated_names = [name for name in dir(_new_module) if not name.startswith("_")]


def __getattr__(name):
    if name in _deprecated_names:
        _warn_deprecation(
            f"The `{_old_module_name}` module is deprecated. Use `{_new_module_name}` instead.\n"
            f"To fix this warning, change `{_old_module_name}.{name} to `{_new_module_name}.{name}`."
        )
        return getattr(_new_module, name)

    raise AttributeError(f"module {__name__} has no attribute {name}")


def __dir__():
    return dir(_new_module)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial, reduce
from typing import Optional, Tuple, Callable

import numpy as np

import jax
import netket.jax as nkjax
from jax import numpy as jnp
from jax.tree_util import (
    tree_flatten,
    tree_unflatten,
    tree_map,
    tree_leaves,
)

from netket.utils import random_seed, mpi
from netket.utils.mpi import MPI_jax_comm
from netket.utils.types import PyTree, PRNGKeyT, SeedT, Scalar


def tree_ravel(pytree: PyTree) -> Tuple[jnp.ndarray, Callable]:
    """Ravel (i.e. flatten) a pytree of arrays down to a 1D array.

    Args:
      pytree: a pytree to ravel

    Returns:
      A pair where the first element is a 1D array representing the flattened and
      concatenated leaf values, and the second element is a callable for
      unflattening a 1D vector of the same length back to a pytree of of the same
      structure as the input ``pytree``.
    """
    leaves, treedef = tree_flatten(pytree)
    flat, unravel_list = nkjax.vjp(_ravel_list, *leaves)
    unravel_pytree = lambda flat: tree_unflatten(treedef, unravel_list(flat))
    return flat, unravel_pytree


def _ravel_list(*lst):
    return jnp.concatenate([jnp.ravel(elt) for elt in lst]) if lst else jnp.array([])


def eval_shape(fun, *args, has_aux=False, **kwargs):
    """
    Returns the dtype of forward_fn(pars, v)
    """
    if has_aux:
        out, _ = jax.eval_shape(fun, *args, **kwargs)
    else:
        out = jax.eval_shape(fun, *args, **kwargs)
    return out


def tree_size(tree: PyTree) -> int:
    """
    Returns the sum of the size of all leaves in the tree.
    It's equivalent to the number of scalars in the pytree.
    """
    return sum(tree_leaves(tree_map(lambda x: x.size, tree)))


def is_complex(x):
    """
    Returns True if x has a complex dtype
    """
    return jnp.issubdtype(x.dtype, jnp.complexfloating)


def is_real(x):
    """
    Returns True if x has a floating point real dtype
    """
    return jnp.issubdtype(x.dtype, jnp.floating)


def tree_leaf_iscomplex(pars: PyTree) -> bool:
    """
    Returns true if at least one leaf in the tree has complex dtype.
    """
    return any(jax.tree_leaves(jax.tree_map(is_complex, pars)))


def tree_leaf_isreal(pars: PyTree) -> bool:
    """
    Returns true if at least one leaf in the tree has real dtype.
    """
    return any(jax.tree_leaves(jax.tree_map(is_real, pars)))


def is_complex_dtype(typ):
    """
    Returns True if typ is a complex dtype
    """
    return jnp.issubdtype(typ, jnp.complexfloating)


def is_real_dtype(typ):
    """
    Returns True if typ is a floating real dtype
    """
    return jnp.issubdtype(typ, jnp.floating)


# Return the type holding the real part of the input type
def dtype_real(typ):
    """
    If typ is a complex dtype returns the real counterpart of typ
    (eg complex64 -> float32, complex128 ->float64).
    Returns typ otherwise.
    """
    if np.issubdtype(typ, np.complexfloating):
        if typ == np.dtype("complex64"):
            return np.dtype("float32")
        elif typ == np.dtype("complex128"):
            return np.dtype("float64")
        else:
            raise TypeError("Unknown complex floating type {}".format(typ))
    else:
        return typ


def tree_ishomogeneous(pars: PyTree) -> bool:
    """
    Returns true if all leaves have real dtype or all leaves have complex dtype.
    """
    return not (tree_leaf_isreal(pars) and tree_leaf_iscomplex(pars))


def dtype_complex(typ):
    """
    Return the complex dtype corresponding to the type passed in.
    If it is already complex, do nothing
    """
    if is_complex_dtype(typ):
        return typ
    elif typ == np.dtype("float32"):
        return np.dtype("complex64")
    elif typ == np.dtype("float64"):
        return np.dtype("complex128")
    else:
        raise TypeError("Unknown complex type for {}".format(typ))


def maybe_promote_to_complex(*types):
    """
    Maybe promotes the first argument to it's complex counterpart given by
    dtype_complex(typ) if any of the arguments is complex
    """
    main_typ = types[0]

    for typ in types:
        if is_complex_dtype(typ):
            return dtype_complex(main_typ)
    else:
        return main_typ


def tree_conj(t: PyTree) -> PyTree:
    r"""
    Conjugate all complex leaves. The real leaves are left untouched.
    Args:
        t: pytree
    """
    return jax.tree_map(lambda x: jax.lax.conj(x) if jnp.iscomplexobj(x) else x, t)


def tree_dot(a: PyTree, b: PyTree) -> Scalar:
    r"""
    compute the dot product of two pytrees

    Args:
        a, b: pytrees with the same treedef

    Returns:
        A scalar equal the dot product of of the flattened arrays of a and b.
    """
    return jax.tree_util.tree_reduce(
        jax.numpy.add,
        jax.tree_map(jax.numpy.sum, jax.tree_multimap(jax.numpy.multiply, a, b)),
    )


def tree_cast(x: PyTree, target: PyTree) -> PyTree:
    r"""
    cast x the types of target

    Args:
        x: a pytree with arrays as leaves
        target: a pytree with the same treedef as x
                where only the dtypes of the leaves are accessed
    Returns:
        A pytree where each leaf of x is cast to the dtype of the corresponding leaf in target.
        The imaginary part of complex leaves which are cast to real is discarded.
    """
    # astype alone would also work, however that raises ComplexWarning when casting complex to real
    # therefore the real is taken first where needed
    return jax.tree_multimap(
        lambda x, target: (x if jnp.iscomplexobj(target) else x.real).astype(
            target.dtype
        ),
        x,
        target,
    )


def tree_axpy(a: Scalar, x: PyTree, y: PyTree) -> PyTree:
    r"""
    compute a * x + y

    Args:
      a: scalar
      x, y: pytrees with the same treedef
    Returns:
        The sum of the respective leaves of the two pytrees x and y
        where the leaves of x are first scaled with a.
    """
    return jax.tree_multimap(lambda x_, y_: a * x_ + y_, x, y)


def _to_real(x):
    if jnp.iscomplexobj(x):
        return x.real, x.imag
        # TODO find a way to make it a nop?
        # return jax.vmap(lambda y: jnp.array((y.real, y.imag)))(x)
    else:
        return x


def _tree_to_real(x):
    return jax.tree_map(_to_real, x)


# invert the transformation using linear_transpose (AD)
def _tree_reassemble_complex(x, target, fun=_tree_to_real):
    (res,) = jax.linear_transpose(fun, target)(x)
    return tree_conj(res)


def tree_to_real(pytree: PyTree) -> Tuple[PyTree, Callable]:
    """Replace all complex leaves of a pytree with a tuple of 2 real leaves.

    Args:
      pytree: a pytree to convert to real

    Returns:
      A pair where the first element is the converted real pytree,
      and the second element is a callable for converting back a real pytree
      to a complex pytree of of the same structure as the input pytree.
    """
    return _tree_to_real(pytree), partial(
        _tree_reassemble_complex, target=pytree, fun=_tree_to_real
    )


class HashablePartial(partial):
    """
    A class behaving like functools.partial, but that retains it's hash
    if it's created with a lexically equivalent (the same) function and
    with the same partially applied arguments and keywords.

    It also stores the computed hash for faster hashing.
    """

    def __init__(self, *args, **kwargs):
        self._hash = None

    def __eq__(self, other):
        return (
            type(other) is HashablePartial
            and self.func.__code__ == other.func.__code__
            and self.args == other.args
            and self.keywords == other.keywords
        )

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(
                (self.func.__code__, self.args, frozenset(self.keywords.items()))
            )

        return self._hash

    def __repr__(self):
        return f"<hashable partial {self.func.__name__} with args={self.args} and kwargs={self.keywords}, hash={hash(self)}>"


# jax.tree_util.register_pytree_node(
#    HashablePartial,
#    lambda partial_: ((), (partial_.func, partial_.args, partial_.keywords)),
#    lambda args, _: StaticPartial(args[0], *args[1], **args[2]),
# )


def compose(*funcs):
    """
    function composition

    compose(f,g,h)(x) is equivalent to f(g(h(x)))
    """

    def _compose(f, g):
        return lambda *args, **kwargs: f(g(*args, **kwargs))

    return reduce(_compose, funcs)


def PRNGKey(seed: Optional[SeedT] = None, root: int = 0, comm=MPI_jax_comm) -> PRNGKeyT:
    """
    Initialises a PRNGKey using an optional starting seed.
    The same seed will be distributed to all processes.
    """
    if seed is None:
        key = jax.random.PRNGKey(random_seed())
    elif isinstance(seed, int):
        key = jax.random.PRNGKey(seed)
    else:
        key = seed

    key, _ = mpi.mpi_bcast_jax(key, root=root, comm=comm)

    return key


def mpi_split(key, root=0, comm=MPI_jax_comm) -> PRNGKeyT:
    """
    Split a key across MPI nodes in the communicator.
    Only the input key on the root process matters.

    Arguments:
        key: The key to split. Only considered the one on the root process.
        root: (default=0) The root rank from which to take the input key.
        comm: (default=MPI.COMM_WORLD) The MPI communicator.

    Returns:
        A PRNGKey depending on rank number and key.
    """

    # Maybe add error/warning if in_key is not the same
    # on all MPI nodes?
    keys = jax.random.split(key, mpi.n_nodes)

    keys, _ = mpi.mpi_bcast_jax(keys, root=root)

    return keys[mpi.rank]


class PRNGSeq:
    """
    A sequence of PRNG keys genrated based on an initial key.
    """

    def __init__(self, base_key: Optional[SeedT] = None):
        if base_key is None:
            base_key = PRNGKey()
        elif isinstance(base_key, int):
            base_key = PRNGKey(base_key)
        self._current = base_key

    def __iter__(self):
        return self

    def __next__(self):
        self._current = jax.random.split(self._current, num=1)[0]
        return self._current

    def next(self):
        return self.__next__()

    def take(self, num: int):
        """
        Returns an array of `num` PRNG keys and advances the iterator accordingly.
        """
        keys = jax.random.split(self._current, num=num + 1)
        self._current = keys[-1]
        return keys[:-1]



# Copyright 2020 The Netket Authors. - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from netket import legacy as nk
from generate_data import generate
import sys
import numpy as np


mpi_rank = nk.MPI.rank()
nk.random.seed(123)

# Generate and load the data
N = 10
hi, rotations, training_samples, training_bases, ha, psi = generate(
    N, n_basis=2 * N, n_shots=500, seed=1234
)

# Machine
ma = nk.machine.RbmSpinPhase(hilbert=hi, alpha=1)
ma.init_random_parameters(seed=1234, sigma=0.01)

# Sampler
sa = nk.sampler.MetropolisLocal(machine=ma, n_chains=32)

# Optimizer
op = nk.optimizer.AdaDelta()

# Quantum State Reconstruction
qst = nk.Qsr(
    sampler=sa,
    optimizer=op,
    samples=training_samples,
    rotations=rotations,
    bases=training_bases,
    n_samples=1000,
    n_samples_data=1000,
    sr=None,
)


qst.add_observable(ha, "Energy")


def run():
    for step in qst.iter(500, 50):
        obs = qst.get_observable_stats()
        if mpi_rank == 0:
            print("step={}".format(step))
            print("observables={}".format(obs))

            # Compute fidelity with exact state
            psima = ma.to_array(normalize=True)
            fidelity = np.abs(np.vdot(psima, psi))
            print("fidelity={}".format(fidelity))

            # Compute NLL on training data
            nll = qst.nll(
                rotations=rotations,
                samples=training_samples,
                bases=training_bases,
                log_norm=ma.log_norm(),
            )
            print("negative log likelihood={}".format(nll))

            # Print output to the console immediately
            sys.stdout.flush()

            # Save current parameters to file
            ma.save("test.wf")



from typing import Tuple, Union

from .abstract_hilbert import AbstractHilbert


class AbstractParticle(AbstractHilbert):
    """Abstract class for the Hilbert space of particles
    in continuous space.
    This class defines the common interface that
    can be used to interact with particles defined
    in continuous space.
    """

    def __init__(
        self, N: int, L: Tuple[float, ...], pbc: Union[bool, Tuple[bool, ...]]
    ):
        """
        Constructs new ``Particles`` given specifications
         of the continuous space they are defined in.

        Args:
            N: Number of particles
            L: spatial extension in each spatial dimension
            pbc: Whether or not to use periodic boundary
                conditions for each spatial dimension.
                If bool, its value will be used for all spatial
                dimensions.
        """
        self._N = N
        self._L = L

        if isinstance(pbc, bool):
            pbc = [pbc] * len(self._L)

        self._pbc = pbc
        if not len(self._L) == len(self._pbc):
            raise ValueError(
                """`pbc` must be either a bool or a tuple indicating the periodicity of each spatial dimension."""
            )

        super().__init__()

    @property
    def size(self) -> int:
        return self._N * len(self._L)

    @property
    def n_particles(self) -> int:
        r"""The number of particles"""
        return self._N

    @property
    def extend(self) -> Tuple[float, ...]:
        r"""Spatial extension in each spatial dimension"""
        return self._L

    @property
    def pbc(self) -> Tuple[bool, ...]:
        r"""Whether or not to use periodic boundary conditions for each spatial dimension"""
        return self._pbc

    @property
    def _attrs(self):
        return (
            self.size,
            self.extend,
            self.pbc,
        )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import jax.numpy as jnp
import numpy as np

import netket as nk

from .. import common  # noqa: F401

SEED = 111


@pytest.fixture(
    params=[pytest.param(M, id=f"Fock(M={M})") for M in [0, 2, 3, 4, 5, 6, 8]]
)
def vstate(request):
    M = request.param
    # keep this a prime number so we get different sizes on every rank...
    hi = nk.hilbert.Fock(M, 1)

    ma = nk.models.RBM(
        alpha=1,
        dtype=float,
        hidden_bias_init=nk.nn.initializers.normal(),
        visible_bias_init=nk.nn.initializers.normal(),
    )

    return nk.vqs.MCState(
        nk.sampler.MetropolisLocal(hi),
        ma,
    )


@pytest.mark.parametrize("normalize", [True, False])
def test_to_array(vstate, normalize):
    psi = vstate.to_array(normalize=normalize)

    if normalize:
        np.testing.assert_allclose(jnp.linalg.norm(psi), 1.0)

    psi_norm = psi / jnp.linalg.norm(psi)

    assert psi.shape == (vstate.hilbert.n_states,)

    x = vstate.hilbert.all_states()
    psi_exact = jnp.exp(vstate.log_value(x))
    psi_exact = psi_exact / jnp.linalg.norm(psi_exact)

    np.testing.assert_allclose(psi_norm, psi_exact)


@pytest.fixture(
    params=[pytest.param(M, id=f"Fock(M={M})") for M in [0, 2, 3, 4, 5, 6, 8]]
)
def vstate_rho(request):
    M = request.param
    # keep this a prime number so we get different sizes on every rank...
    hi = nk.hilbert.Fock(M, 1)

    ma = nk.models.NDM()

    return nk.vqs.MCMixedState(
        nk.sampler.MetropolisLocal(nk.hilbert.DoubledHilbert(hi)),
        ma,
    )


@pytest.mark.parametrize("normalize", [True, False])
def test_to_matrix(vstate_rho, normalize):
    rho = vstate_rho.to_matrix(normalize=normalize)

    if normalize:
        np.testing.assert_allclose(jnp.trace(rho), 1.0)

    rho_norm = rho / jnp.trace(rho)

    assert rho.shape == (
        vstate_rho.hilbert.physical.n_states,
        vstate_rho.hilbert.physical.n_states,
    )

    x = vstate_rho.hilbert.all_states()
    rho_exact = jnp.exp(vstate_rho.log_value(x)).reshape(rho.shape)
    rho_exact = rho_exact / jnp.trace(rho_exact)

    np.testing.assert_allclose(rho_norm, rho_exact)
